#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3

from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.rollbackChangeRoute53WeightsStep import RollbackChangeRoute53WeightsStep
from ecs_crd.updateCanaryReleaseInfoStep import UpdateCanaryReleaseInfoStep

class ChangeRoute53WeightsStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos, 'Change Route 53 Weights', logger)

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            strategy = self._consume_strategy()
            self._change_weights(strategy)
            return CheckGreenHealthStep(self.infos, self.logger)
        except Exception as e:
            self.logger.error('ChangeRoute53WeightsStep', exc_info=True)
            self.infos.exit_exception = e
            self.infos.exit_code = 5
            return RollbackChangeRoute53WeightsStep(self.infos, self.logger)

    def _change_weights(self, strategy):
        """update balancing ratio beetween blue's and green's DNS"""
        # by default 100 % for green / 0% for blue
        green_weight, blue_weight = (100, 0)
        # if exist canary strategy , set weights
        if (strategy):
            green_weight = strategy.weight
            blue_weight = 100 - green_weight

        self.logger.info('')
        self._log_information(key='Dns weight blue', value=f"{blue_weight}%")
        self._log_information(key='Dns weight green', value=f"{green_weight}%")
        self.logger.info('')

        client = boto3.client('route53')
        for item in self.infos.fqdn:
            self._change_weights_by_fqdn(item, client, blue_weight, green_weight)

        if (green_weight == 100):
            self.infos.strategy_infos.clear()
        self.logger.info('')
        self._wait(strategy.wait, "Changing DNS's Weights")

    def _change_weights_by_fqdn(self, fqdn, client, blue_weight, green_weight):
        self._log_information(key='Fqdn', value=fqdn.name)
        client.change_resource_record_sets(
            HostedZoneId = fqdn.hosted_zone_id,
            ChangeBatch={
                'Comment': 'Alter Route53 records sets for canary blue-green deployment',
                'Changes': [
                    {
                        'Action': 'UPSERT',
                        'ResourceRecordSet': {
                            'Name': f"{fqdn.name}.",
                            'Type': 'CNAME',
                            'SetIdentifier': self.infos.blue_infos.canary_release,
                            'Weight': blue_weight,
                            'TTL': 60,
                            'ResourceRecords': [
                                {
                                    'Value': self.infos.blue_infos.alb_dns
                                },
                            ]
                        },
                    },
                    {
                        'Action': 'UPSERT',
                        'ResourceRecordSet': {
                            'Name': f"{fqdn.name}.",
                            'Type': 'CNAME',
                            'SetIdentifier': self.infos.green_infos.canary_release,
                            'Weight': green_weight,
                            'TTL': 60,
                            'ResourceRecords': [
                                {
                                    'Value': self.infos.green_infos.alb_dns
                                },
                            ]
                        },
                    }
                ]
            }
        )
       

    def _consume_strategy(self):
        """consume the first strategy of the canary release's definition"""
        result = None
        if self.infos.strategy_infos:
            result, tmp = (self.infos.strategy_infos[0],
                           self.infos.strategy_infos[1:])
            self.infos.strategy_infos = tmp
        return result


class CheckGreenHealthStep(CanaryReleaseDeployStep):
    """ Check health status of Green's """

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        self._nb_max_initial_test = 4
        self._nb_initial_test = 1
        super().__init__(infos, 'Check Health Green LoadBalancer', logger)

    def _find_health_checks(self):
        """return list of state of health check load balancer"""
        result = []
        client = boto3.client('elbv2', region_name=self.infos.region)
        targetGroupArns = self._find_target_group_arns()
        for e in targetGroupArns:
            response = client.describe_target_health(TargetGroupArn=e['OutputValue'])
            state = "UNKNOWN"
            if response['TargetHealthDescriptions']:
                state = response['TargetHealthDescriptions'][0]['TargetHealth']['State'].upper()
            self.logger.info('')
            self._log_information(key='Target Group', value=e['OutputKey'][:-3])
            self._log_information(key='Arn', value=e['OutputValue'])
            self._log_information(key='State', value=state)
            result.append(state)
        return result

    def _is_all_full_states(self, health_checks, states):
        if not states:
            return False
        return len(list(filter(lambda x: x in states, health_checks))) == len(health_checks)

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            while True:
                health_checks = self._find_health_checks()
                is_full_healthy = self._is_all_full_states(health_checks, ['HEALTHY'])
                if is_full_healthy:
                    break
                is_healthy_and_initial = self._is_all_full_states(health_checks, ['HEALTHY', 'INITIAL'])
                if is_healthy_and_initial:
                    if self._nb_initial_test < self._nb_max_initial_test:
                        self._wait(15, f'Waiting for service to start (attempts {self._nb_initial_test}/{self._nb_max_initial_test})')
                        self._nb_initial_test += 1
                        continue
                raise ValueError(f'Invalid state for Green TargetGroup')
            # all health check is ok
            if self.infos.strategy_infos:
                return ChangeRoute53WeightsStep(self.infos, self.logger)
            else:
                return UpdateCanaryReleaseInfoStep(self.infos, self.logger)

        except Exception as e:
            self.logger.error('CheckGreenHealthStep', exc_info=True)
            self.infos.exit_exception = e
            self.infos.exit_code = 14
            return RollbackChangeRoute53WeightsStep(self.infos, self.logger)

    def _find_target_group_arns(self):
        client = boto3.client('cloudformation', region_name=self.infos.region)
        response = client.describe_stacks(StackName= self.infos.green_infos.stack_name)
        return  filter(lambda x: x['OutputKey'].startswith('TargetGroup'),response['Stacks'][0]['Outputs'])
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3
import json
import time

from abc import ABC, abstractmethod
from ecs_crd.defaultJSONEncoder import DefaultJSONEncoder
from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep

class CreateStackStep(CanaryReleaseDeployStep):

    def __init__(self, infos, title, logger, stack_infos):
        """initializes a new instance of the class"""
        super().__init__(
            infos, 
            title, 
            logger
        )
        self.timer = 5
        self.stack_infos = stack_infos

    @abstractmethod
    def _on_success(self):
        pass
        
    @abstractmethod
    def _on_fail(self):
        pass
    
    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            self.logger.info('')
            self._log_information(key='Stack Name',
                                  value=self.stack_infos.stack_name)
            self.logger.info('')
            self.logger.info(f'Creating stack in progress ...')
            client = boto3.client('cloudformation',
                                  region_name=self.infos.region)
            self._create_stack(client)
            self._monitor(client)
            self.infos.save()
            return self._on_success()
        except Exception as e:
            self.logger.error('CreateGreenStackStep', exc_info=True)
            self.infos.exit_exception = e
            self.infos.exit_code = 12
            return self._on_fail()

    def _monitor(self, client):
        """pause the process and wait for the result of the cloud formation stack creation"""
        wait = 0
        while True:
            wait += self.timer
            w = self._second_to_string(wait)
            self.logger.info('')
            time.sleep(self.timer)
            self.logger.info(f'Creating stack in progress ... [{w} elapsed]')
            response = client.describe_stacks(StackName=self.stack_infos.stack_id)
            stack = response['Stacks'][0]
            response2 = client.list_stack_resources(StackName=self.stack_infos.stack_id)
            for resource in response2['StackResourceSummaries']:
                message = resource['LogicalResourceId'].ljust(40, '.') + \
                          resource['ResourceStatus']
                if 'ResourceStatusReason' in resource:
                    message += f' ( {resource["ResourceStatusReason"]} )'
                self.logger.info(message)

            if stack['StackStatus'] == 'CREATE_IN_PROGRESS':
                continue
            elif stack['StackStatus'] == 'CREATE_COMPLETE':
                break
            else:
                raise ValueError('Error creation green cloudformation stack')

    def _create_stack(self, client):
        if self.stack_infos.stack_id:
            return 
        """created the cloud formation stack"""
        payload = json.dumps(self.stack_infos.stack, cls=DefaultJSONEncoder)
        response = client.create_stack(
            StackName = self.stack_infos.stack_name,
            TemplateBody = payload,
            OnFailure='DELETE',
            Capabilities=['CAPABILITY_NAMED_IAM'],
            Tags=[
                {
                    'Key': 'Project',
                    'Value': self.infos.project
                },
                {
                    'Key': 'Service',
                    'Value': self.infos.service_name
                },
                {
                    'Key': 'Environment',
                    'Value': self.infos.environment
                },
                {
                    'Key': 'Version',
                    'Value': self.infos.service_version
                }]
            )
        self.stack_infos.stack_id = response['StackId']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3

from ecs_crd.finishDeploymentStep import FinishDeploymentStep
from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep

class SendNotificationBySnsStep(CanaryReleaseDeployStep):
    
    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos, 'Send Notification', logger, with_end_log=True)

    def _find_sns_topic_notifcation(self):
        if 'sns_topic_notifications' in self.configuration['canary']:
            result = self.configuration['canary']['sns_topic_notifications']
            if self.infos.exit_code == 0 and 'on_success' in  result:
                return self._bind_data(result['on_success'])
            if self.infos.exit_code != 0 and 'on_fail' in  result:
                return self._bind_data(result['on_fail'])
            return None

    def _on_execute(self):
        if self.infos.action != 'validate':
            sns_topic_notification = self._find_sns_topic_notifcation()
            if sns_topic_notification:
                message = f'ECS canary {self.infos.action} of the "{self.infos.service_name}" service was successful.'
                if self.infos.exit_code != 0:
                    message = f'ECS canary {self.infos.action} of the "{self.infos.service_name}" service was failed.'
                message +=f'\nAccount        : {self.infos.account_id}'
                message +=f'\nRegion         : {self.infos.region}'
                message +=f'\nEnvironment    : {self.infos.environment}'
                message +=f'\nCluster        : {self.infos.cluster_name}'
                message +=f'\nProject        : {self.infos.project}'
                message +=f'\nService        : {self.infos.service_name}'
                message +=f'\nVersion        : {self.infos.service_version}'
                fqdn = ''
                for item in self.infos.fqdn:
                    fqdn += item.name + ','
                message +=f'\nFqdn           : {fqdn}'
                message +=f'\nExit Code      : {self.infos.exit_code}'
                message +=f'\nMessage        : {self.infos.exit_exception}'
                subject = '[SUCCESS]' if self.infos.exit_code == 0 else '[FAIL]'
                subject += f' {self.infos.action} - {self.infos.service_name}' 
                client = boto3.client('sns', region_name=self.infos.region)
                self._log_information(key='Sending Notification ... ', value=None)
                try:
                    client.publish(
                        TopicArn = sns_topic_notification,
                        Message = message,
                        Subject = subject
                    )
                    self._log_information(key='Notification sended with success.', value=None)
                except Exception as e:
                    if self.infos.exit_code == 0:
                        self.infos.exit_code = 100
                        self.infos.exit_exception = e
                    self.logger.error(self.title, exc_info=True)
            else:
                self._log_information(key='No notification to send', value=None)
        return FinishDeploymentStep(self.infos, self.logger)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import boto3

from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.prepareDeploymentIamPoliciesStep import PrepareDeploymentIamPoliciesStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep

class PrepareDeploymentListenersStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos,  f'Prepare {infos.action}( Listeners )', logger)
        self.current_priority = None

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            cfn = self.infos.green_infos.stack['Resources']['Service']['Properties']['LoadBalancers']
            for item in cfn:
                target_group = self.infos.green_infos.stack['Resources'][item['TargetGroupArn']['Ref']]
                listener_infos = self._find_listener_infos(target_group)
                #TODO à revoir
                ttt = target_group['Properties']['Tags'] 
                container_name = next((x for x in ttt if x['Key'] == 'Container'), None)['Value']
                container_port = str(next((x for x in ttt if x['Key'] == 'ContainerPort'), None)['Value'])

                if not listener_infos:
                    container_name = next((x for x in ttt if x['Key'] == 'Container'), None)['Value']
                    raise ValueError('Not found listener informations for container "{}:{}", target group port "{}"'.format(container_name, container_port, target_group['Properties']['Port'])) 

                listener_rule_infos = self._find_listener_rule_infos(listener_infos)
                # not exist listener rule , then create a new listener
                if not listener_rule_infos:
                    self._process_listener(listener_infos, item, target_group, container_name, container_port)
                else:
                # else create listener rules
                    self._process_listener_rules(listener_rule_infos, item, target_group, container_name, container_port)

            self.infos.save()
            return PrepareDeploymentIamPoliciesStep(self.infos, self.logger)

        except Exception as e:
            self.infos.exit_code = 8
            self.infos.exit_exception = e
            self.logger.error(self.title, exc_info=True)
            return SendNotificationBySnsStep(self.infos, self.logger)

    def _process_listener_rules(self, listener_rule_infos, item, target_group, container_name, container_port):
        """convert listener dto informations to cloudformation List of Application Load Balancer Listener rules"""
        host_port = str(self._find_host_port(container_name, container_port))
        host_port = '(dynamic)' if host_port == '0' else host_port
        self._log_sub_title(f'Container "{container_name}:{container_port}"')
        self._log_information(key="Arn",value=listener_rule_infos.listener_arn, indent=1)
        self._log_information(key="Port", value=listener_rule_infos.configuration['port'], indent=1)
        self._log_information(key="Rules", value='', indent=1)
        # rules
        count = 1
        for rule in listener_rule_infos.configuration['rules']:
            resource_key = item['TargetGroupArn']['Ref'].replace('TargetGroup', f'ListenerRule{count}')
            self.infos.green_infos.stack['Resources'][resource_key] = self._convert_2_listener_rule(listener_rule_infos, item, rule)
            count +=1
        # certificates
        certificates = self._find_certificates(listener_rule_infos.configuration)
        if certificates:
            self._log_sub_title('Certificates')
            listener_certificate = {}
            listener_certificate['Type'] = 'AWS::ElasticLoadBalancingV2::ListenerCertificate'
            listener_certificate['Properties'] = {}
            listener_certificate['Properties']['Certificates'] = []
            for cert in certificates:
                certificate = {}
                certificate['CertificateArn'] = cert['CertificateArn']
                listener_certificate['Properties']['Certificates'].append(certificate)
                self._log_information(key='- DomainName', value=cert['DomainName'], indent=1)
                self._log_information(key='  ARN', value=cert['CertificateArn'], indent=1)
            listener_certificate['Properties']['ListenerArn'] = listener_rule_infos.listener_arn
            self.infos.green_infos.stack['Resources']['ListenerCertificate'] = listener_certificate

    def _process_listener(self, listener_infos, item, target_group, container_name, container_port):
        """convert listener dto informations to cloudformation Application Load Balancer Listener"""
        # Resource
        listener = {}
        listener['Type'] = "AWS::ElasticLoadBalancingV2::Listener"
        listener['Properties'] = {}
        listener['Properties']['Port'] = listener_infos['port']

        self._log_sub_title(f'Container "{container_name}:{container_port}"')

        if 'protocol' in listener_infos:
            listener['Properties']['Protocol'] = listener_infos['protocol'].upper()
        else:
            listener['Properties']['Protocol'] = 'HTTP'

        if listener['Properties']['Protocol'] == 'HTTPS':
            listener['Properties']['SslPolicy'] = 'ELBSecurityPolicy-2016-08'
            listener['Properties']['Certificates'] = []
            for cert in self._find_certificates(listener_infos):
                certificate = {}
                certificate['CertificateArn'] = cert['CertificateArn']
                listener['Properties']['Certificates'].append(certificate)

        listener['Properties']['LoadBalancerArn'] = {}
        listener['Properties']['LoadBalancerArn']['Ref'] = "LoadBalancer"

        listener['Properties']['DefaultActions'] = []
        action = {}
        action['Type'] = 'forward'
        action['TargetGroupArn'] = {}
        action['TargetGroupArn']['Ref'] = item['TargetGroupArn']['Ref']
        listener['Properties']['DefaultActions'].append(action)

        host_port = str(self._find_host_port(container_name, container_port))
        host_port = '(dynamic)' if host_port == '0' else host_port
        self._log_information(key="Port", value=listener_infos['port'], indent=1)
        cfn_resource_listener_name = item['TargetGroupArn']['Ref'].replace('TargetGroup', 'Listener')
        self.infos.green_infos.stack['Resources'][cfn_resource_listener_name] = listener

    def _find_listener_rule_infos(self, listener_infos):
        """check if the AWS Application Load Balancer Listerner exist"""
        listener = None
        client = boto3.client('elbv2', region_name=self.infos.region)
        response = client.describe_listeners(LoadBalancerArn=self.infos.green_infos.alb_arn)
        for item in response['Listeners']:
            if int(item['Port']) == int(listener_infos['port']):
                listener = item
                break
        if listener:
            for item in self.infos.listener_rules_infos:
                if item.configuration == listener_infos:
                    item.listener_arn = listener['ListenerArn']
                    item.current_priority = self._find_last_listener_rule_priority(item)    
                    return item

    def _find_certificates(self, listener_infos):
        """find all cerficates used by listener"""
        result = []
        if 'certificates' in listener_infos:
            client = boto3.client('acm', region_name=self.infos.region)
            response = client.list_certificates()
            for cert in response['CertificateSummaryList']:
                for certificate in listener_infos['certificates']:
                    if cert['DomainName'] == self._bind_data(certificate):
                        result.append(cert)
        return result

    def _find_listener_infos(self, target_group):
        """find listener information in configuration"""
        t = target_group['Properties']['Tags']
        container_name = list(filter(lambda x: x['Key'] == 'Container', t))[0]['Value']
        container_port = int(list(filter(lambda x: x['Key'] == 'ContainerPort', t))[0]['Value'])

        for item in self.configuration['listeners']:
            port = int(item['target_group']['container']['port'])
            if container_port == port:
                listener_container_name = 'default'
                if 'name' in item['target_group']['container']:
                    listener_container_name = item['target_group']['container']['name']
                if listener_container_name.lower() == container_name.lower():
                    return item
        return None

    def _convert_2_listener_rule(self, listener_rule_infos, item, rule):
        listener_rule = {}
        listener_rule['Type'] = "AWS::ElasticLoadBalancingV2::ListenerRule"
        listener_rule['Properties'] = {}
        listener_rule['Properties']['Priority'] = self._calculate_avalaible_priority_rule(listener_rule_infos, rule)
        listener_rule['Properties']['ListenerArn'] = listener_rule_infos.listener_arn
        self._log_information(key="- Priority",value=str(listener_rule['Properties']['Priority']), indent=2)
        # actions
        listener_rule['Properties']['Actions'] = []
        self._log_information(key="Actions",value='', indent=4)
        order = 1
        add_default_action = True
        if 'actions' in rule:
            for action in rule['actions']:
                action, is_default_action = self._convert_2_action(action, order)
                listener_rule['Properties']['Actions'].append(action)
                if is_default_action:
                    add_default_action = False
                order +=1
        # add default forward action to container target group
        if add_default_action:
            action, is_default_action = self._convert_2_action_forward(item, order)
            listener_rule['Properties']['Actions'].append(action)
        
        # conditions
        listener_rule['Properties']['Conditions'] = []
        if 'conditions' in rule:
            self._log_information(key="Conditions",value='', indent=4)
            for condition in rule['conditions']:
                listener_rule['Properties']['Conditions'].append(self._convert_2_condition(condition))
        return listener_rule
    
    def _find_last_listener_rule_priority(self, listener_rule_infos):
        client = boto3.client('elbv2', region_name = self.infos.region)
        response = client.describe_rules(ListenerArn = listener_rule_infos.listener_arn)
        priorities = [2]
        for item in response['Rules']:
            priority = self._to_int(item['Priority'])
            if priority:
                priorities.append(priority)
        return max(priorities, key=lambda x:x) + 1
        
    def _find_host_port(self, container_name, container_port):
        """ find the host port for the tupe container name / container port"""
        cfn_container_definitions = self.infos.green_infos.stack['Resources']['TaskDefinition']['Properties']['ContainerDefinitions']
        container_info = next((x for x in cfn_container_definitions if x['Name'] == container_name), None)
        return next((x for x in container_info['PortMappings'] if str(x['ContainerPort']) == str(container_port)), None)['HostPort']

    def _calculate_avalaible_priority_rule(self, listener_rule_infos, rule):
        """calculate avalaible priority rule"""
        result = None
        if 'priority' in rule:
            return rule['priority']
        else:
            result = listener_rule_infos.current_priority
            listener_rule_infos.current_priority += 1
            return result

    def _convert_2_condition(self, item):
        condition = {}
        self._process_property(
            source = item,
            target = condition,
            pattern= 'host-header|http-header|http-request-method|path-pattern|source-ip',
            source_property = 'field',
            indent = 5,
            multi = True
        )
        property_name = None
        if condition['Field'] == 'host-header':
            property_name = 'HostHeaderConfig'
        elif condition['Field'] == 'http-header':
            property_name = 'HttpHeaderConfig'
        elif condition['Field'] == 'http-request-method':
            property_name = 'HttpRequestMethodConfig'
        elif condition['Field'] == 'path-pattern':
            property_name = 'PathPatternConfig'
        else:
            property_name = 'SourceIpConfig'

        condition[property_name] = {}
        self._log_information(key=property_name,value='',indent=7)
        condition[property_name]['Values'] = []
        self._log_information(key='Values',value='',indent=9)
        for v in item['values']:
            data = self._bind_data(v)
            condition[property_name]['Values'].append(data)
            self._log_information(key='- ' +data,value=None,indent=11)

        return condition

    def _convert_2_action_forward(self, item, order):
        action = {}
        action['Type'] = 'forward'
        action['Order'] = order
        action['TargetGroupArn'] = {}
        action['TargetGroupArn']['Ref'] = item['TargetGroupArn']['Ref']
        self._log_information(key=f'- Type', value = action['Type'], indent=5)
        self._log_information(key=f'Order', value = action['Order'], indent=7)
        self._log_information(key='TargetGroupArn', value ='-> container target group created', indent=7)
        return action, True

    def _convert_2_action(self, item, order):
        action = {}
        is_default_action = False
        # action type
        self._process_property(
            source = item,
            target = action,
            pattern= 'authenticate-oidc|authenticate-cognito|fixed-response|redirect',
            source_property = 'type',
            indent = 5,
            multi = True
        )
        action['Order'] = order if 'order' not in item else item['order']
        self._log_information(key='Order',value=action['Order'],indent=7)
        property_name = None
        if action['Type'] == 'authenticate-oidc':
            property_name = 'AuthenticateOidcConfig'
        elif action['Type'] == 'authenticate-cognito':
            property_name = 'AuthenticateCognitoConfig'
        elif action['Type'] == 'fixed-response':
            property_name = 'FixedResponseConfig'
            is_default_action = True
        else:
            property_name = 'RedirectConfig'
            is_default_action = True
        
        if property_name:
            action[property_name]= {}
            self._log_information(key=property_name,value='',indent=7)
            for k in item['config'].keys():
                key = self._to_pascal_case(k)
                src = item['config'][k]
                if isinstance(src, str):
                    action[property_name][key] = self._bind_data(src)
                else:
                    action[property_name][key] = src
                self._log_information(key=key,value=action[property_name][key],indent=9)
        
        return action, is_default_action



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3

from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep
from ecs_crd.destroyInitStackStep import DestroyInitStackStep
from ecs_crd.destroyStackStep import DestroyStackStep
from ecs_crd.canaryReleaseInfos import LoadBalancerInfos

class DestroyBlueStackStep(DestroyStackStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(
            infos, 
            'Delete Blue Cloudformation Stack', 
            logger,
            infos.blue_infos
        )

    def _on_success(self):
        if self.infos.action == 'deploy':
            return SendNotificationBySnsStep(self.infos, self.logger)
        if self.infos.action == 'undeploy':
            #self._clean_route_53_record_sets()
            return DestroyInitStackStep(self.infos, self.logger)
        
    def _on_fail(self):
        return None
   
    def _clean_route_53_record_sets(self):
        """remove CNAME in route 53"""
        client = boto3.client('route53')
        client.change_resource_record_sets(
            HostedZoneId=self.infos.hosted_zone_id,
            ChangeBatch={
                'Comment': 'Remove Route53 records sets for canary blue-green deployment',
                'Changes': [
                    {
                        'Action': 'DELETE',
                        'ResourceRecordSet': {
                            'Name': f"{self.infos.fqdn}.",
                            'Type': 'CNAME',
                            'SetIdentifier': self.infos.blue_infos.canary_release,
                            'TTL': 60,
                            'ResourceRecords': [
                                {
                                    'Value': self.infos.blue_infos.alb_dns
                                },
                            ]
                        },
                    },
                    {
                        'Action': 'DELETE',
                        'ResourceRecordSet': {
                            'Name': f"{self.infos.fqdn}.",
                            'Type': 'CNAME',
                            'SetIdentifier': self.infos.green_infos.canary_release,
                            'TTL': 60,
                            'ResourceRecords': [
                                {
                                    'Value': self.infos.green_infos.alb_dns
                                },
                            ]
                        },
                    }
                ]
            }
        )
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.prepareDeploymentListenersStep import PrepareDeploymentListenersStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep

class PrepareDeploymentTargetGroupsStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initialize a new instance of class"""
        super().__init__(infos, f'Prepare {infos.action} ( Target groups )', logger)

    def _process_target_group_tags(self, item, target_group):
        """update tags informations for the target group"""
        target_group['Properties']['Tags'] = []

        tag = {}
        tag['Key'] = "Environment"
        tag['Value'] = self.infos.environment
        target_group['Properties']['Tags'].append(tag)

        tag = {}
        tag['Key'] = "Project"
        tag['Value'] = self.infos.project
        target_group['Properties']['Tags'].append(tag)

        tag = {}
        tag['Key'] = "Service"
        tag['Value'] = self.infos.service_name
        target_group['Properties']['Tags'].append(tag)

        tag = {}
        tag['Key'] = "Version"
        tag['Value'] = self.infos.service_version
        target_group['Properties']['Tags'].append(tag)

        tag = {}
        tag['Key'] = "Container"
        tag['Value'] = item['ContainerName']
        target_group['Properties']['Tags'].append(tag)

        tag = {}
        tag['Key'] = "ContainerPort"
        tag['Value'] = str(item['ContainerPort'])
        target_group['Properties']['Tags'].append(tag)

        tag = {}
        tag['Key'] = "CanaryRelease"
        tag['Value'] = self.infos.green_infos.canary_release
        target_group['Properties']['Tags'].append(tag)

    def _process_target_group_port(self, item, target_group_info, target_group):
        """update port informations for the target group"""
        if 'port' in target_group_info:
            if isinstance(target_group_info['port'],int):
                target_group['Properties']['Port'] = int(
                    target_group_info['port'])
            else:
                if 'green' in target_group_info['port'] and 'blue' in target_group_info['port']:
                    target_group['Properties']['Port'] = int(
                        target_group_info['port'][self.infos.elected_release])
                else:
                    raise ValueError('Not found port target group informations for container {}:{} '.format(
                        item['ContainerName'], item['ContainerPort']))
        else:
            target_group['Properties']['Port'] = int(item['ContainerPort'])
        if target_group['Properties']['Port'] < 0:
            raise ValueError('{} is invalid for the port of target group'.format(
                target_group['Properties']['Port']))
        self._log_information(
            key='Port', value=target_group['Properties']['Port'], indent=1)

    def _process_target_group_protocol(self, item, target_group_info, target_group):
        """update protocol informations for the target group"""
        if 'protocol' in target_group_info:
            target_group['Properties']['Protocol'] = target_group_info['protocol'].upper()
        else:
            target_group['Properties']['Protocol'] = 'HTTP'
        if target_group['Properties']['Protocol'] not in ['HTTP', 'HTTPS']:
            raise ValueError('{} is not valid protocle'.format(
                target_group['Properties']['Protocol']))
        self._log_information(
            key='Procotol', value=target_group['Properties']['Protocol'], indent=1)

    def _process_target_group_attributes(self, item, target_group_info, target_group):
        """update attributes informations for the target group"""
        if 'target_group_attributes' in target_group_info:
            target_group['TargetGroupAttributes'] = []
            self._log_information(
                key='Target group attributes', value='', indent=1)
            for e in target_group_info['target_group_attributes']:
                target_group_attribute = {}
                target_group_attribute['Key'] = str(e['key'])
                target_group_attribute['Value'] = str(e['value'])
                target_group['TargetGroupAttributes'].append(
                    target_group_attribute)
                self._log_information(
                    key='- '+target_group_attribute['Key'], value=target_group_attribute['Value'], indent=2)

    def _process_target_group_health_check(self, item, target_group_info, target_group):
        """update health check informations for the target group"""
        if 'health_check' not in target_group_info:
            raise ValueError('health_check is mandatory.')
        target_group['Properties']['HealthCheckEnabled'] = "true"
        health_check_infos = target_group_info['health_check']
        self._log_information(key='Health Check', value='', indent=1)
        host_port = self._find_host_port(
            item['ContainerName'], item['ContainerPort'])
        if host_port != 0:
            # Port
            if 'port' in health_check_infos:
                if isinstance(health_check_infos['port'],int):
                    target_group['Properties']['HealthCheckPort'] = int(
                        health_check_infos['port'])
                else:
                    if 'green' in health_check_infos['port'] and 'blue' in health_check_infos['port']:
                        target_group['Properties']['HealthCheckPort'] = int(
                            health_check_infos['port'][self.infos.elected_release])
            self._log_information(
                key='Host port', value=target_group['Properties']['HealthCheckPort'], indent=2)
        else:
            self._log_information(key='Host port', value='dynamic', indent=2)
        # Interval seconds
        if 'interval_seconds' in health_check_infos:
            target_group['Properties']['HealthCheckIntervalSeconds'] = int(
                health_check_infos['interval_seconds'])
            self._log_information(
                key='Interval Seconds', value=target_group['Properties']['HealthCheckIntervalSeconds'], indent=2)

        # Healthy threshold count
        if 'healthy_threshold_count' in health_check_infos:
            target_group['Properties']['HealthyThresholdCount'] = int(
                health_check_infos['healthy_threshold_count'])
            self._log_information(key='Healthy Threshold Count',
                                  value=target_group['Properties']['HealthyThresholdCount'], indent=2)

        # Unhealthy threshold count
        if 'unhealthy_threshold_count' in health_check_infos:
            target_group['Properties']['UnhealthyThresholdCount'] = int(
                health_check_infos['unhealthy_threshold_count'])
            self._log_information(key='Unhealthy Threshold Count',
                                  value=target_group['Properties']['UnhealthyThresholdCount'], indent=2)

        # Path
        target_group['Properties']['HealthCheckPath'] = '/'
        if 'path' in health_check_infos:
            target_group['Properties']['HealthCheckPath'] = health_check_infos['path']

        self._log_information(
            key='Path', value=target_group['Properties']['HealthCheckPath'], indent=2)

        # Protocol
        if 'protocol' in health_check_infos:
            target_group['Properties']['HealthCheckProtocol'] = health_check_infos['protocol'].upper()
        else:
            target_group['Properties']['HealthCheckProtocol'] = 'HTTP'

        self._log_information(
            key='Protocol', value=target_group['Properties']['HealthCheckProtocol'], indent=2)

        # Matcher
        matcher = {}
        matcher['HttpCode'] = "200"
        if 'matcher' in health_check_infos:
            matcher['HttpCode'] = health_check_infos['matcher']

        target_group['Properties']['Matcher'] = matcher
        self._log_information(
            key='Matcher', value=matcher['HttpCode'], indent=2)

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            cfn = self.infos.green_infos.stack['Resources']['Service']['Properties']['LoadBalancers']
            for item in cfn:

                target_group = {}
                target_group['Type'] = "AWS::ElasticLoadBalancingV2::TargetGroup"
                target_group['Properties'] = {}
                target_group['Properties']['Name'] = (
                    '{}-{}'.format(self.infos.id[:10], item['TargetGroupArn']['Ref'].replace('TargetGroup', '')[:18]))+'-tg'
                target_group['Properties']['VpcId'] = self.infos.vpc_id

                self._process_target_group_tags(item, target_group)

                target_group_info = None
                for elmt in self.configuration['target_groups']:
                    container_name = 'default'
                    if 'name' in elmt['container']:
                        container_name = elmt['container']['name']
                    if (elmt['container']['port'] == item['ContainerPort']
                            and container_name == item['ContainerName']):
                        target_group_info = elmt
                        break

                if target_group_info == None:
                    raise ValueError('Not found target group informations for container {}:{} '.format(
                        item['ContainerName'], item['ContainerPort']))

                container_name = 'default' if 'name' not in target_group_info[
                    'container'] else target_group_info['container']['name']
                self._log_sub_title('Container "{}:{}"'.format(
                    container_name, target_group_info['container']['port']))
                self._process_target_group_port(
                    item, target_group_info, target_group)
                self._process_target_group_protocol(
                    item, target_group_info, target_group)
                self._process_target_group_attributes(
                    item, target_group_info, target_group)
                self._process_target_group_health_check(
                    item, target_group_info, target_group)
                self.infos.green_infos.stack['Resources'][item['TargetGroupArn']
                                                          ['Ref']] = target_group

                # add result output
                self._add_to_output_cloud_formation(item['TargetGroupArn']['Ref'])

            self.infos.save()

            return PrepareDeploymentListenersStep(self.infos, self.logger)

        except Exception as e:
            self.infos.exit_code = 7
            self.infos.exit_exception = e
            self.logger.error(self.title, exc_info=True)
            return SendNotificationBySnsStep(self.infos, self.logger)

    def _add_to_output_cloud_formation(self, target_group_name):
        output = self.infos.green_infos.stack['Outputs']
        id = target_group_name+'Arn'
        output[id] = {}
        output[id]['Description']=f'The ARN of {target_group_name}'
        output[id]['Value']={}
        output[id]['Value']['Ref'] = target_group_name

    def _find_host_port(self, container_name, container_port):
        """find the host port for tuple container name/ container port """
        cfn_container_definitions = self.infos.green_infos.stack['Resources'][
            'TaskDefinition']['Properties']['ContainerDefinitions']
        container_info = next(
            (x for x in cfn_container_definitions if x['Name'] == container_name), None)
        return next((x for x in container_info['PortMappings'] if x['ContainerPort'] == container_port), None)['HostPort']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3

from ecs_crd.createStackStep import CreateStackStep
from ecs_crd.createGreenStackStep import CreateGreenStackStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep

class CreateInitStackStep(CreateStackStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(
            infos,
            'Create Init Cloudformation Stack', 
            logger,
            infos.init_infos)
        #if self.infos.init_infos.stack: 
  
    def _on_success(self):
        return CreateGreenStackStep(self.infos, self.logger)
        
    def _on_fail(self):
       return SendNotificationBySnsStep(self.infos, self.logger)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3
import json
import time
import traceback

from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.destroyGreenStackStep import DestroyGreenStackStep
from ecs_crd.applyStrategyStep import CheckGreenHealthStep
from ecs_crd.destroyGreenStackStep import DestroyGreenStackStep
from ecs_crd.defaultJSONEncoder import DefaultJSONEncoder

class ScaleUpServiceStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos,'Scale Up Service', logger)

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            service_arn = self._find_service_arn()
            self.logger.info('')
            self._log_information(key='Service', value=self.infos.green_infos.stack_name)
            self._log_information(key='Arn', value=service_arn)
            self.logger.info('')
            self.logger.info('Scaling up in progress ...')
            self.logger.info('')
            client = boto3.client('ecs', region_name=self.infos.region)
            if self.infos.scale_infos.desired > 1:
                client.update_service(
                    cluster=self.infos.cluster,
                    service=service_arn,
                    desiredCount=self.infos.scale_infos.desired,
                    deploymentConfiguration={
                        'maximumPercent': 100 * self.infos.scale_infos.desired,
                        'minimumHealthyPercent': 100
                    },
                    forceNewDeployment= True

                )
            else:
                client.update_service(
                    cluster=self.infos.cluster,
                    service=service_arn,
                    desiredCount=self.infos.scale_infos.desired,
                    forceNewDeployment= True)
            self._wait(self.infos.scale_infos.wait, 'Scaling up in progress')
            self.logger.info('')
            self.logger.info(f'Desired instances : {self.infos.scale_infos.desired}')
            return CheckGreenHealthStep(self.infos, self.logger)

        except Exception as e:
            self.logger.error('ScaleUpServiceStep', exc_info=True)
            self.infos.exit_exception = e
            self.infos.exit_code = 13
            return DestroyGreenStackStep(self.infos, self.logger)

    def _find_service_arn(self):
        """find AWS ARN of service"""
        client = boto3.client('cloudformation', region_name=self.infos.region)
        response = client.describe_stacks(StackName= self.infos.green_infos.stack_name)
        output = next(x for x in response['Stacks'][0]['Outputs'] if x['OutputKey']=='ServiceArn')
        return output['OutputValue']


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.prepareDeploymentTargetGroupsStep import PrepareDeploymentTargetGroupsStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep

class PrepareDeploymentTaskDefinitionStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos, f'Prepare {infos.action} ( Task definition )', logger)

    def _process_cpu(self, item, cfn):
        if 'cpu' in item:
            cfn['Cpu'] = int(item['cpu'])
            self._log_information(key='Cpu', value=cfn["Cpu"], ljust=10, indent=1)

    def _process_memory(self, item, cfn):
        if 'memory' in item:
            cfn['Memory'] = int(item['memory'])
            self._log_information(key='Memory', value=cfn["Memory"], ljust=10, indent=1)

    def _process_network_mode(self, item, cfn):
        if 'network_mode' in item:
            cfn['NetworkMode'] = item['network_mode']
            self._log_information(key='NetworkMode', value=cfn["NetworkMode"], ljust=10, indent=1)

    def _process_pid_mode(self, item, cfn):
        if 'pid_mode' in item:
            cfn['PidMode'] = item['pid_mode']
            self._log_information(key='PidMode', value=cfn["PidMode"], ljust=10, indent=1)

    def _process_ipc_mode(self, item, cfn):
        if 'ipc_mode' in item:
            cfn['IpcMode'] = item['ipc_cmode']
            self._log_information(key='IpcMode', value=cfn["IpcMode"], ljust=10, indent=1)

    def _process_requires_compatibilities(self, item, cfn):
        if 'requires_compatibilities' in item:
            self._log_information(key='Requires Compatibilities', value='', ljust=10, indent=1)
            cfn['RequiresCompatibilities'] =[]
            for e in item['requires_compatibilities']:
                cfn['RequiresCompatibilities'].append(e)
                self._log_information(key='- '+e, value=None, ljust=10, indent=1)

    def _process_volumes(self, item, cfn):
        if 'volumes' in item:
            cfn['Volumes'] = []
            self._log_information(key='Volumes', value=None, ljust=10, indent=1)
            for e in item['volumes']:
                volume = {}
                volume['Name'] = self._bind_data(item['name'])
                self._log_information(key='- Name', value=volume['Name'], ljust=10, indent=2)
                if 'docker_volume_configuration' in e:
                    volume['DockerVolumeConfiguration'] = {}
                    self._process_docker_volume_configuration(e['docker_volume_configuration'], volume['DockerVolumeConfiguration'])
                if 'host' in e:
                    volume['Host'] = {}
                    self._process_host(e['host'],volume['Host'])
                cfn['Volumes'].append(volume)

    def _process_host(self, item, cfn):
        self._log_information(key='Host', value=None, ljust=10, indent=4)
        if 'source_path' in item:
            cfn['SourcePath'] = self._bind_data(item['source_path'])
            self._log_information(key='SourcePath', value=cfn['SourcePath'], ljust=10, indent=5)

    def _process_docker_volume_configuration(self, item, cfn):
        self._log_information(key='DockerVolumeConfiguration', value=None, ljust=10, indent=4)
        if 'autoprovision' in item:
            cfn['Autoprovision'] = str(item['autoprovision'])
            self._log_information(key='Autoprovision', value=cfn['Autoprovision'], ljust=10, indent=5)
        if 'driver' in item:
            cfn['Driver'] = str(item['driver'])
            self._log_information(key='Driver', value=cfn['Driver'], ljust=10, indent=5)
        if 'driver_opts' in item:
            cfn['DriverOpts'] = []
            self._log_information(key='DriverOpts', value=None, ljust=10, indent=5)
            # TODO fix
            for e in item['driver_opts']:
                elmt = {}
                key = None
                value = None
                for a in e.keys():
                    key = a
                for a in elmt.values():
                    value = str(a)
                elmt[key] = value
                cfn['DriverOpts'].append(elmt)
                self._log_information(key='- '+key, value=value, ljust=10, indent=6)
        if 'labels' in item:
            self._log_information(key='Labels', value=None, ljust=10, indent=5)
            cfn['Labels'] = []
            for e in item['labels']:
                elmt = {}
                key = None
                value = None
                for a in e.keys():
                    key = a
                for a in elmt.values():
                    value = str(a)
                elmt[key] = self._bind_data(value)
                cfn['Labels'].append(elmt)
                self._log_information(key='- '+key, value=value, ljust=10, indent=6)
        if 'scope' in item:
            cfn['Scope'] = item['scope']
            self._log_information(key='Scope', value=cfn['Scope'], ljust=10, indent=5)

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            cfn = self.infos.green_infos.stack['Resources']['TaskDefinition']['Properties']
            item = self.configuration['service']
            self._process_cpu(item, cfn)
            self._process_memory(item, cfn)
            self._process_network_mode(item, cfn)
            self._process_ipc_mode(item, cfn)
            self._process_pid_mode(item, cfn)
            self._process_requires_compatibilities(item, cfn)
            self._process_volumes(item, cfn)
            self.infos.save()
            return PrepareDeploymentTargetGroupsStep(self.infos, self.logger)         

        except Exception as e:
            self.infos.exit_code = 6
            self.infos.exit_exception = e
            self.logger.error(self.title, exc_info=True)
            return SendNotificationBySnsStep(self.infos, self.logger)




import pytest
from unittest.mock import MagicMock
import logging

from ecs_crd.canaryReleaseInfos import CanaryReleaseInfos
from ecs_crd.prepareDeploymentServiceDefinitionStep import PrepareDeploymentServiceDefinitionStep
from ecs_crd.canaryReleaseInfos import ScaleInfos

logger = logging.Logger('mock')
infos = CanaryReleaseInfos(action='test')
step = PrepareDeploymentServiceDefinitionStep(infos, logger)

def test_process_step_scaling_policy_configuration_adjustment_type_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['adjustment_type']='a'
        target = {}
        step._process_step_scaling_policy_configuration_adjustment_type(source, target)

def test_process_step_scaling_policy_configuration_adjustment_type_valid():
    source = {}
    source['adjustment_type']='ChangeInCapacity'
    target = {}
    step._process_step_scaling_policy_configuration_adjustment_type(source, target)
    assert target['AdjustmentType'] == source['adjustment_type']

    source['adjustment_type']='PercentChangeInCapacity'
    target = {}
    step._process_step_scaling_policy_configuration_adjustment_type(source, target)
    assert target['AdjustmentType'] == source['adjustment_type']

    source['adjustment_type']='ExactCapacity'
    target = {}
    step._process_step_scaling_policy_configuration_adjustment_type(source, target)
    assert target['AdjustmentType'] == source['adjustment_type']

def test_process_step_scaling_policy_configuration_cooldown_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['cooldown']='a'
        target = {}
        step._process_step_scaling_policy_configuration_cooldown(source, target)

def test_process_step_scaling_policy_configuration_cooldown_valid():
    source = {}
    source['cooldown']= 60
    target = {}
    step._process_step_scaling_policy_configuration_cooldown(source, target)
    assert target['Cooldown'] == source['cooldown']

def test_process_step_scaling_policy_configuration_metric_aggregation_type_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['metric_aggregation_type']='a'
        target = {}
        step._process_step_scaling_policy_configuration_metric_aggregation_type(source, target)

def test_process_step_scaling_policy_configuration_metric_aggregation_type_valid():
    source = {}
    
    source['metric_aggregation_type']='Average'
    target = {}
    step._process_step_scaling_policy_configuration_metric_aggregation_type(source, target)
    assert target['MetricAggregationType'] == source['metric_aggregation_type']

    source['metric_aggregation_type']='Minimum'
    target = {}
    step._process_step_scaling_policy_configuration_metric_aggregation_type(source, target)
    assert target['MetricAggregationType'] == source['metric_aggregation_type']

    source['metric_aggregation_type']='Maximum'
    target = {}
    step._process_step_scaling_policy_configuration_metric_aggregation_type(source, target)
    assert target['MetricAggregationType'] == source['metric_aggregation_type']


def test_process_step_scaling_policy_configuration_step_adjustments_metric_interval_lower_bound_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['metric_interval_lower_bound']='a'
        target = {}
        step._process_step_scaling_policy_configuration_step_adjustments_metric_interval_lower_bound(source, target)

def test_process_step_scaling_policy_configuration_step_adjustments_metric_interval_lower_bound_valid():
    source = {}
    source['metric_interval_lower_bound']= 0
    target = {}
    step._process_step_scaling_policy_configuration_step_adjustments_metric_interval_lower_bound(source, target)
    assert target['MetricIntervalLowerBound'] == source['metric_interval_lower_bound']

def test_process_process_step_scaling_policy_configuration_step_adjustments_metric_interval_upper_bound_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['metric_interval_upper_bound']='a'
        target = {}
        step._process_step_scaling_policy_configuration_step_adjustments_metric_interval_upper_bound(source, target)

def test_process_process_step_scaling_policy_configuration_step_adjustments_metric_interval_upper_bound_valid():
    source = {}
    source['metric_interval_upper_bound']= 0
    target = {}
    step._process_step_scaling_policy_configuration_step_adjustments_metric_interval_upper_bound(source, target)
    assert target['MetricIntervalUpperBound'] == source['metric_interval_upper_bound']

def test_process_process_step_scaling_policy_configuration_step_adjustments_scaling_adjustment_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['scaling_adjustment']='a'
        target = {}
        step._process_step_scaling_policy_configuration_step_adjustments_scaling_adjustment(source, target)

def test_process_process_step_scaling_policy_configuration_step_adjustments_scaling_adjustment_valid():
    source = {}
    source['scaling_adjustment']= 0
    target = {}
    step._process_step_scaling_policy_configuration_step_adjustments_scaling_adjustment(source, target)
    assert target['ScalingAdjustment'] == source['scaling_adjustment']

def test_process_cloudwatch_alarm_metric_name_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['metric_name']='a'
        target = {}
        step._process_cloudwatch_alarm_metric_name(source, target)

def test_process_cloudwatch_alarm_metric_name_empty():
    with pytest.raises(ValueError):
        source = {}
        target = {}
        step._process_cloudwatch_alarm_metric_name(source, target)

def test_process_cloudwatch_alarm_metric_name_valid():
    source = {}
    source['metric_name']='CPUUtilization'
    target = {}
    step._process_cloudwatch_alarm_metric_name(source, target)
    assert target['MetricName'] == source['metric_name']

    source['metric_name']='MemoryUtilization'
    target = {}
    step._process_cloudwatch_alarm_metric_name(source, target)
    assert target['MetricName'] == source['metric_name']

def test_process_cloudwatch_alarm_alarm_description_empty():
    source = {}
    target = {}
    target['MetricName'] = 'CPUUtilization'
    step._process_cloudwatch_alarm_alarm_description(source, target)
    assert target['AlarmDescription'] == f'Containers {target["MetricName"]} High'

def test_process_cloudwatch_alarm_alarm_description_not_empty():
    source = {}
    target = {}
    target['MetricName'] = 'CPUUtilization'
    source['alarm_description']='test'
    step._process_cloudwatch_alarm_alarm_description(source, target)
    assert target['AlarmDescription'] == source['alarm_description']

def test_process_cloudwatch_alarm_namespace():
    source = {}
    expected= {}
    source['namespace'] = 'test'
    step._process_cloudwatch_alarm_namespace(source, expected)
    assert expected['Namespace'] == source['namespace']

def test_process_cloudwatch_alarm_statistic_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['statistic']='a'
        target = {}
        step._process_cloudwatch_alarm_statistic(source, target)

def test_process_cloudwatch_alarm_statistic_valid():
    source = {}
    target = {}

    step._process_cloudwatch_alarm_statistic(source, target)
    assert target['Statistic'] == 'Average'
    
    source['statistic']='Average'
    step._process_cloudwatch_alarm_statistic(source, target)
    assert target['Statistic'] == source['statistic']

    source['statistic']='Minimum'
    target = {}
    step._process_cloudwatch_alarm_statistic(source, target)
    assert target['Statistic'] == source['statistic']

    source['statistic']='Maximum'
    target = {}
    step._process_cloudwatch_alarm_statistic(source, target)
    assert target['Statistic'] == source['statistic']

    source['statistic']='SampleCount'
    target = {}
    step._process_cloudwatch_alarm_statistic(source, target)
    assert target['Statistic'] == source['statistic']

    source['statistic']='Sum'
    target = {}
    step._process_cloudwatch_alarm_statistic(source, target)
    assert target['Statistic'] == source['statistic']

def test_process_cloudwatch_alarm_period_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['period']='a'
        target = {}
        step._process_cloudwatch_alarm_period(source, target)

def test_process_cloudwatch_alarm_period_valid():
    source = {}
    source['period']= 60
    target = {}
    step._process_cloudwatch_alarm_period(source, target)
    assert target['Period'] == source['period']

def test_process_cloudwatch_alarm_evaluation_periods_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['evaluation_periods']='a'
        target = {}
        step._process_cloudwatch_alarm_evaluation_periods(source, target)

def test_process_cloudwatch_alarm_evaluation_periods_valid():
    source = {}
    source['evaluation_periods']= 2
    target = {}
    step._process_cloudwatch_alarm_evaluation_periods(source, target)
    assert target['EvaluationPeriods'] == source['evaluation_periods']

def test_process_cloudwatch_alarm_evaluation_threshold_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['threshold']='a'
        target = {}
        step._process_cloudwatch_alarm_threshold(source, target)

def test_process_cloudwatch_alarm_evaluation_threshold_required():
    with pytest.raises(ValueError):
        source = {}
        target = {}
        step._process_cloudwatch_alarm_threshold(source, target)

def test_process_cloudwatch_alarm_evaluation_periods_valid():
    source = {}
    source['threshold']= 50
    target = {}
    step._process_cloudwatch_alarm_threshold(source, target)
    assert target['Threshold'] == source['threshold']

def test_process_cloudwatch_alarm_comparison_operator_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['comparison_operator']='a'
        target = {}
        step._process_cloudwatch_alarm_comparison_operator(source, target)

def test_process_cloudwatch_alarm_comparison_operator_required():
    with pytest.raises(ValueError):
        source = {}
        target = {}
        step._process_cloudwatch_alarm_comparison_operator(source, target)

def test_process_cloudwatch_alarm_comparison_operator_valid():
    source = {}
    target = {}
    
    source['comparison_operator']='GreaterThanOrEqualToThreshold'
    step._process_cloudwatch_alarm_comparison_operator(source, target)
    assert target['ComparisonOperator'] == source['comparison_operator']

    source['comparison_operator']='GreaterThanThreshold'
    step._process_cloudwatch_alarm_comparison_operator(source, target)
    assert target['ComparisonOperator'] == source['comparison_operator']

    source['comparison_operator']='LessThanOrEqualToThreshold'
    step._process_cloudwatch_alarm_comparison_operator(source, target)
    assert target['ComparisonOperator'] == source['comparison_operator']

    source['comparison_operator']='LessThanThreshold'
    step._process_cloudwatch_alarm_comparison_operator(source, target)
    assert target['ComparisonOperator'] == source['comparison_operator']

def test_process_scheduling_strategy_invalid():
    with pytest.raises(ValueError):
        source = {}
        source['scheduling_strategy']='a'
        target = {}
        step._process_scheduling_strategy(source, target)

def test_process_scheduling_strategy_valid():
    source = {}
    target = {}
    
    source['scheduling_strategy']='DAEMON'
    step._process_scheduling_strategy(source, target)
    assert target['SchedulingStrategy'] == source['scheduling_strategy']

    source['scheduling_strategy']='REPLICA'
    step._process_scheduling_strategy(source, target)
    assert target['SchedulingStrategy'] == source['scheduling_strategy']

def test_process_application_autoscaling_scalable_target_min_capacity_valid():
    source = {}
    target = {}
    
    # default value
    step.infos.scale_infos = ScaleInfos()
    step.infos.scale_infos.desired = 2
    step._process_application_autoscaling_scalable_target_min_capacity(source, target)
    assert target['MinCapacity'] == step.infos.scale_infos.desired 
    
    # set valude
    source['min_capacity'] = 1
    step._process_application_autoscaling_scalable_target_min_capacity(source, target)
    assert target['MinCapacity'] == source['min_capacity']  

def test_process_application_autoscaling_scalable_target_max_capacity_invalid():
    source = {}
    source['max_capacity'] = 'a'
    target = {}
    with pytest.raises(ValueError):
        step._process_application_autoscaling_scalable_target_max_capacity(source, target)

def test_process_application_autoscaling_scalable_target_max_capacity_valid():
    source = {}
    target = {}
    
    # default value
    step.infos.scale_infos = ScaleInfos()
    step.infos.scale_infos.desired = 2
    step._process_application_autoscaling_scalable_target_max_capacity(source, target)
    assert target['MaxCapacity'] == step.infos.scale_infos.desired 
    
    # set valude
    source['max_capacity'] = 1
    step._process_application_autoscaling_scalable_target_max_capacity(source, target)
    assert target['MaxCapacity'] == source['max_capacity']  

def test_process_application_autoscaling_scalable_target_role_arn_valid():
    source = {}
    target = {}
    source['role_arn'] = 'a'
    step._process_application_autoscaling_scalable_target_role_arn(source, target)
    assert target['RoleARN'] == source['role_arn']

def test_process_application_autoscaling_scalable_target_role_arn_required():
    source = {}
    target = {}
    with pytest.raises(ValueError):
        step._process_application_autoscaling_scalable_target_role_arn(source, target)

def test_process_application_auto_scaling_scaling_policy_policy_type_invalid():
    source = {}
    target = {}
    with pytest.raises(ValueError):
        source['policy_type']='a'
        step._process_application_auto_scaling_scaling_policy_policy_type(source, target)
   
def test_process_application_auto_scaling_scaling_policy_policy_type_valid():
    source = {}
    target = {}
    
    source['policy_type']='SimpleScaling'
    step._process_application_auto_scaling_scaling_policy_policy_type(source, target)
    assert target['PolicyType'] == source['policy_type']

    source['policy_type']='StepScaling'
    step._process_application_auto_scaling_scaling_policy_policy_type(source, target)
    assert target['PolicyType'] == source['policy_type']

    source['policy_type']='TargetTrackingScaling'
    step._process_application_auto_scaling_scaling_policy_policy_type(source, target)
    assert target['PolicyType'] == source['policy_type']

def test_process_placement_stategies_strategy_field_required():
    source = {}
    target = {}
    with pytest.raises(ValueError):
        step._process_placement_stategies_strategy_field(source, target)

def test_process_placement_stategies_strategy_field_valid():
    source = {}
    source['field'] ='test'
    target = {}
    step._process_placement_stategies_strategy_field(source, target)
    assert target['Field'] == source['field']

def test_process_placement_stategies_strategy_type_required():
    source = {}
    target = {}
    with pytest.raises(ValueError):
        step._process_placement_stategies_strategy_type(source, target)

def test_process_placement_stategies_strategy_field_invalid():
    source = {}
    source['type']='test'
    target = {}
    with pytest.raises(ValueError):
        step._process_placement_stategies_strategy_type(source, target)

def test_process_placement_stategies_strategy_field_valid():
    source = {}
    source['type']='binpack'
    target = {}
    step._process_placement_stategies_strategy_type(source, target)
    assert target['Type'] == source['type']

    source['type']='random'
    target = {}
    step._process_placement_stategies_strategy_type(source, target)
    assert target['Type'] == source['type']

    source['type']='spread'
    target = {}
    step._process_placement_stategies_strategy_type(source, target)
    assert target['Type'] == source['type']
    
def test_process_placement_constraints_contraint_type_required():
    source = {}
    target = {}
    with pytest.raises(ValueError):
        step._process_placement_constraints_contraint_type(source, target)

def test_process_placement_stategies_strategy_type_invalid():
    source = {}
    source['type']='test'
    target = {}
    with pytest.raises(ValueError):
        step._process_placement_constraints_contraint_type(source, target)

def test_process_placement_stategies_strategy_type_valid():
    source = {}
    source['type']='distinctInstance'
    target = {}
    step._process_placement_constraints_contraint_type(source, target)
    assert target['Type'] == source['type']

    source['type']='memberOf'
    target = {}
    step._process_placement_constraints_contraint_type(source, target)
    assert target['Type'] == source['type']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.canaryReleaseInfos import StrategyInfos
from ecs_crd.prepareDeploymentInitStackStep import PrepareDeploymentInitStackStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep

class PrepareDeploymentStrategyStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos, f'Prepare {infos.action} ( Canary Strategy )', logger)
        self.min_wait = 40
        self.default_weight = 50
        self.default_wait = 60

    def _process_strategy(self):
        """update strategies informations for the service"""
        source = self.configuration['canary']
        if 'strategy' in source:
            for item in source['strategy']:
                wait = self.default_wait
                if 'wait' in item:
                    wait = int(item['wait'])
                if wait < self.min_wait:
                    wait = self.min_wait
                weight = self.default_weight
                if 'weight' in item:
                    weight = int(item['weight'])
                self.infos.strategy_infos.append(StrategyInfos(weight=weight, wait=wait))

        # no previous stack ( go to 100% )
        if not self.infos.blue_infos.stack_id:
            self.infos.strategy_infos.clear()

        self.infos.strategy_infos.append(StrategyInfos(weight=100, wait=self.default_wait))
        self.infos.strategy_infos = sorted(self.infos.strategy_infos, key=lambda strategy: strategy.weight)        

        for a in self.infos.strategy_infos:
            self._log_information(key='- Weight', value=a.weight, indent=1)
            self._log_information(key='  Wait', value=f'{a.wait}s', indent=1)

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            self._process_strategy()
            return PrepareDeploymentInitStackStep(self.infos,self.logger)
        except Exception as e:
            self.infos.exit_code = 10
            self.infos.exit_exception = e
            self.logger.error(self.title, exc_info=True)
            return SendNotificationBySnsStep(self.infos, self.logger)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep
from ecs_crd.destroyStackStep import DestroyStackStep

class DestroyGreenStackStep(DestroyStackStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(
            infos, 
            'Delete Green Cloudformation Stack', 
            logger,
            infos.green_infos
        )

    def _on_success(self):
        return SendNotificationBySnsStep(self.infos, self.logger)
        
    def _on_fail(self):
        return SendNotificationBySnsStep(self.infos, self.logger)


import pytest
from unittest.mock import MagicMock
import logging

from ecs_crd.canaryReleaseInfos import CanaryReleaseInfos
from ecs_crd.prepareDeploymentContainerDefinitionsStep import PrepareDeploymentContainerDefinitionsStep
from ecs_crd.canaryReleaseInfos import ScaleInfos

logger = logging.Logger('mock')
infos = CanaryReleaseInfos(action='test')
step = PrepareDeploymentContainerDefinitionsStep(infos, logger)

def test_process_container_name_valid():
    # default
    source = {}
    target = {}
    step._process_container_name(source, target)
    target['Name'] == 'default'
    # with name
    source = {}
    source['name']='test'
    target = {}
    step._process_container_name(source, target)
    target['Name'] == source['name']

def test_process_container_name_valid():
    
    # default
    source = {}
    target = {}
    step.infos.account_id='123456789'
    step.infos.region='eu-west-3'
    step.infos.service_name='service'
    step.infos.service_version='latest'

    step._process_container_image(source, target)
    assert target['Image'] == '123456789.dkr.ecr.eu-west-3.amazonaws.com/service:latest'

    # with name
    source = {}
    source['image']='test'
    target = {}
    step._process_container_image(source, target)
    assert target['Image'] == source['image']

def test_process_container_cpu_invalid():
    source = {}
    source['cpu'] = 'a'
    target = {}
    with pytest.raises(ValueError):
        step._process_container_cpu(source, target)

def test_process_container_cpu_valid():
    source = {}
    target = {}
    
    # default value
    step._process_container_cpu(source, target)
    assert target['Cpu'] == 128 
    
    # set value
    source['cpu'] = 256
    target = {}
    step._process_container_cpu(source, target)
    assert target['Cpu'] == source['cpu']  

def test_process_container_entry_point_valid():
     source = {}
     source['entry_point']=[]
     source['entry_point'].append('a')
     source['entry_point'].append('b')
     target = {}
     step._process_container_entry_point(source, target)
     assert target['EntryPoint'] == 'a,b'

def test_process_container_entry_point_invalid():
     source = {}
     source['entry_point']='a'
     target = {}
     with pytest.raises(ValueError):
        step._process_container_entry_point(source, target)

def test_process_container_command_valid():
     source = {}
     source['command']=[]
     source['command'].append('a')
     source['command'].append('b')
     target = {}
     step._process_container_command(source, target)
     assert len(target['Command'])==2
     assert target['Command'][0] == 'a'
     assert target['Command'][1] == 'b'

def test_process_container_command_invalid():
     source = {}
     source['command']='b'
     target = {}
     with pytest.raises(ValueError):
        step._process_container_command(source, target)
   
def test_process_container_dns_search_domains_valid():
     source = {}
     source['dns_search_domains']=[]
     source['dns_search_domains'].append('a')
     source['dns_search_domains'].append('b')
     target = {}
     step._process_container_dns_search_domains(source, target)
     assert len(target['DnsSearchDomains'])==2
     assert target['DnsSearchDomains'][0] == 'a'
     assert target['DnsSearchDomains'][1] == 'b'

def _process_container_dns_search_domains_invalid():
     source = {}
     source['dns_search_domains']='b'
     target = {}
     with pytest.raises(ValueError):
        step._process_container_dns_search_domains(source, target)


def test_process_container_disable_networking_valid():
     source = {}
     source['disable_networking'] = True
     target = {}
     step._process_container_disable_networking(source, target)
     assert target['DisableNetworking'] == source['disable_networking']
     source = {}
     source['disable_networking'] = False
     target = {}
     step._process_container_disable_networking(source, target)
     assert target['DisableNetworking'] == source['disable_networking']

def _process_container_disable_networking_invalid():
     source = {}
     source['disable_networking']='b'
     target = {}
     with pytest.raises(ValueError):
        step._process_container_disable_networking(source, target)

def test_process_container_dns_servers_valid():
     source = {}
     source['dns_servers']=[]
     source['dns_servers'].append('a')
     source['dns_servers'].append('b')
     target = {}
     step._process_container_dns_servers(source, target)
     assert len(target['DnsServers'])==2
     assert target['DnsServers'][0] == 'a'
     assert target['DnsServers'][1] == 'b'

def _process_container_dns_servers_invalid():
     source = {}
     source['dns_servers']='b'
     target = {}
     with pytest.raises(ValueError):
        step._process_container_dns_servers(source, target)
   

def test_process_container_start_timeout_invalid():
    source = {}
    source['start_timeout'] = 'a'
    target = {}
    with pytest.raises(ValueError):
        step._process_container_start_timeout(source, target)

def test_process_container_start_timeout_valid():
    source = {}
    source['start_timeout']=60
    target = {}
    
    step._process_container_start_timeout(source, target)
    assert target['StartTimeout'] ==  source['start_timeout']

def test_process_container_stop_timeout_invalid():
    source = {}
    source['stop_timeout'] = 'a'
    target = {}
    with pytest.raises(ValueError):
        step._process_container_stop_timeout(source, target)

def test_process_container_stop_timeout_valid():
    source = {}
    source['stop_timeout']=60
    target = {}
    
    step._process_container_stop_timeout(source, target)
    assert target['StopTimeout'] == source['stop_timeout']
    
def test_process_container_hostname_valid():
    source = {}
    source['hostname']='a'
    target = {}
    
    step._process_container_hostname(source, target)
    assert target['Hostname'] ==  source['hostname']
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3
import json
import time

from ecs_crd.defaultJSONEncoder import DefaultJSONEncoder
from ecs_crd.createStackStep import CreateStackStep
from ecs_crd.scaleUpServiceStep import ScaleUpServiceStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep


class CreateGreenStackStep(CreateStackStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(
            infos, 
            'Create Green Cloudformation Stack', 
            logger,
            infos.green_infos)
        self.timer = 5

    def _on_success(self):
        return ScaleUpServiceStep(self.infos, self.logger)
        
    def _on_fail(self):
       return SendNotificationBySnsStep(self.infos, self.logger)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3

from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.createInitStackStep import CreateInitStackStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep

class PrepareDeploymentInitStackStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos, f'Prepare {infos.action} ( Init stack )', logger)

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            for item in self.infos.fqdn:
                self.logger.info('')
                self._log_information(key='- Fqdn',value=item.name, indent=1)
                self._log_information(key='HostedZoneName',value=item.hosted_zone_name, indent=3)
                self._log_information(key='HostedZoneId',value=item.hosted_zone_id, indent=3)
            if self.infos.init_infos.stack:
                self._prepare_record_set_group()
            if self.infos.action == 'validate':
                return SendNotificationBySnsStep(self.infos,self.logger)
            else:
                return CreateInitStackStep(self.infos, self.logger)

        except Exception as e:
            self.infos.exit_code = 3
            self.infos.exit_exception = e
            self.logger.error(self.title, exc_info=True)
            return SendNotificationBySnsStep(self.infos, self.logger)

    def _prepare_record_set_group(self):
        count = 1
        target = self.infos.init_infos.stack['Resources']
        for item in self.infos.fqdn:
            self._process_record_set_group_by_fqdn(item, count, target)
            count += 1
        self.infos.save()

    def _process_record_set_group_by_fqdn(self, source, count, target):
        """add RecordSetGroup by fqdn in cloudformation stack"""
        cfn = {}
        cfn['Type'] = 'AWS::Route53::RecordSetGroup'
        properties = {}
        properties['HostedZoneName'] = source.hosted_zone_name.strip('.')+'.'
        properties['RecordSets'] = []
        
        # blue 
        recordset = {}
        recordset['Name'] = source.name
        recordset['Type'] = 'CNAME'
        recordset['TTL'] = {}
        recordset['TTL']['Ref'] = 'Route53RecordetSetTTL'
        recordset['SetIdentifier'] = self.infos.blue_infos.canary_release
        recordset['Weight'] = 0
        recordset['ResourceRecords'] = []
        resource_record = {}
        resource_record['Ref'] = 'LoadBalancerBlue'
        recordset['ResourceRecords'].append(resource_record)
        properties['RecordSets'].append(recordset)

        # green 
        recordset = {}
        recordset['Name'] = source.name
        recordset['Type'] = 'CNAME'
        recordset['TTL'] = {}
        recordset['TTL']['Ref'] = 'Route53RecordetSetTTL'
        recordset['SetIdentifier'] = self.infos.green_infos.canary_release
        recordset['Weight'] = 100
        recordset['ResourceRecords'] = []
        resource_record = {}
        resource_record['Ref'] = 'LoadBalancerGreen'
        recordset['ResourceRecords'].append(resource_record)
        properties['RecordSets'].append(recordset)

        cfn['Properties'] = properties

        target[f'CanaryReleaseRecordSetGroup{count}'] = cfn


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import click
import os
import logging
import boto3
import sys

from ecs_crd.prepareDeploymentGlobalParametersStep import PrepareDeploymentGlobalParametersStep
from ecs_crd.canaryReleaseInfos import CanaryReleaseInfos
from ecs_crd.versionInfos import VersionInfos

# Hum un global meriterait d'etre en MAJ
version_infos = VersionInfos()


class Parameters:
    def __init__(self, logger):
        self.environments = ['dev', 'qua', 'stage', 'preprod', 'prod']
        self.regions = ['us-east-1', 'us-east-2', 'us-west-1', 'us-west-2',
                        'ap-east-1', 'ap-south-1', 'ap-northeast-1',
                        'ap-northeast-2', 'ap-northeast-3', 'ap-southeast-1',
                        'ap-southeast-2', 'ca-central-1', 'cn-north-1',
                        'cn-northwest-1', 'eu-central-1', 'eu-west-1',
                        'eu-west-2', 'eu-west-3', 'eu-north-1', 'sa-east-1',
                        'us-gov-east-1', 'us-gov-west-1']
        self.environment = None
        self.region = None
        self.configuration_file = None
        self.configuration_dir = None
        self.logger = logger

    def validate(self):
        if not self.environment:
            raise ValueError(f'environment is required.')
        if self.environment not in self.environments:
            raise ValueError(f'{self.environment} is not valid environment.')
        if not self.region:
            raise ValueError(f'region is required.')
        if self.region not in self.regions:
            raise ValueError(f'{self.region} is not valid aws region.')

        if not self.configuration_file:
            if not self.configuration_dir or self.configuration_dir == '':
                self.configuration_dir = os.getcwd()
            if not self.configuration_dir.endswith('/'):
                self.configuration_dir += "/"
            if not os.path.exists(self.configuration_dir):
                raise ValueError(f'{self.configuration_dir} not exist.')
            self.configuration_file = f"{self.configuration_dir}{self.environment}.deploy.yml"

        if not os.path.isfile(self.configuration_file):
            raise ValueError(f"{self.configuration_file} doesn't exist.")

        self.logger.info(''.ljust(50, '-'))
        self.logger.info(f'Step : Check parameters')
        self.logger.info(''.ljust(50, '-'))
        self.logger.info(f'Region             : {self.region}')
        self.logger.info(f'Environment        : {self.environment}')
        self.logger.info(f'Configuration      : {self.configuration_file}')
        self.logger.info('')
        self.logger.info(f'Check parameters : COMPLETED')


@click.group()
def main():
    pass

@main.command(help='dry run the configuration file')
@click.option('-e', '--environment', required=True, default='stage', help='Environment to deploy.', show_default=True)
@click.option('-r', '--region', required=True, default='eu-west-3', help='Amazon Web Service region used to deploy ECS service.', show_default=True)
@click.option('-f', '--configuration-file', required=False, help='deployment configuration file.')
@click.option('-d', '--configuration-dir', required=False, help='directory to find the deployment configuration file.')
@click.option('--verbose', is_flag=True, default=False, help='activate verbose log.')
@click.option('--log-file', required=False, help='output log file result.')
def dry_run(
        environment,
        region,
        configuration_file,
        configuration_dir,
        verbose,
        log_file):
    logger, canary_infos = _common_action(environment, region, configuration_file, configuration_dir, verbose, log_file)
    canary_infos.action = 'validate'
    canary_step = PrepareDeploymentGlobalParametersStep(canary_infos, logger)
    while (canary_step):
        canary_step = canary_step.execute()
    sys.exit(canary_infos.exit_code)

@main.command(help='deploy the ECS service')
@click.option('-e', '--environment', required=True, default='stage', help='Environment to deploy.', show_default=True)
@click.option('-r', '--region', required=True, default='eu-west-3', help='Amazon Web Service region used to deploy ECS service.', show_default=True)
@click.option('-f', '--configuration-file', required=False, help='deployment configuration file.')
@click.option('-d', '--configuration-dir', required=False, help='directory to find the deployment configuration file.')
@click.option('--verbose', is_flag=True, default=False, help='activate verbose log.')
@click.option('--log-file', required=False, help='output log file result.')
def deploy(
        environment,
        region,
        configuration_file,
        configuration_dir,
        verbose,
        log_file):
    logger, canary_infos = _common_action(environment, region, configuration_file, configuration_dir, verbose, log_file)
    canary_infos.action = 'deploy'
    canary_step = PrepareDeploymentGlobalParametersStep(canary_infos, logger)
    while (canary_step):
        canary_step = canary_step.execute()
    sys.exit(canary_infos.exit_code)

def _common_action(environment, region, configuration_file, configuration_dir, verbose, log_file):
    logger = _create_logger(verbose, log_file)
    parameters = Parameters(logger)
    parameters.environment = environment
    parameters.region = region
    parameters.configuration_file = configuration_file
    parameters.configuration_dir = configuration_dir
    parameters.validate()
    canary_infos = CanaryReleaseInfos(
        environment = parameters.environment,
        region = parameters.region,
        configuration_file = parameters.configuration_file,
        ecs_crd_version = version_infos.version
    )
    return logger, canary_infos

@main.command(help='undeploy the ECS service')
@click.option('-e', '--environment', required=True, default='stage', help='Environment to deploy.', show_default=True)
@click.option('-r', '--region', required=True, default='eu-west-3', help='Amazon Web Service region used to deploy ECS service.', show_default=True)
@click.option('-f', '--configuration-file', required=False, help='deployment configuration file.')
@click.option('-d', '--configuration-dir', required=False, help='directory to find the deployment configuration file.')
@click.option('--verbose', is_flag=True, default=False, help='activate verbose log.')
@click.option('--log-file', required=False, help='output log file result.')
def undeploy(
        environment,
        region,
        configuration_file,
        configuration_dir,
        verbose,
        log_file):
    logger, canary_infos = _common_action(environment, region, configuration_file, configuration_dir, verbose, log_file)
    canary_infos.action = 'undeploy'
    canary_step = PrepareDeploymentGlobalParametersStep(canary_infos, logger)
    while (canary_step):
        canary_step = canary_step.execute()
    sys.exit(canary_infos.exit_code)

@main.command(name='version', help='show CLI version')
def version():
    version_infos = VersionInfos()
    print(f'{version_infos.description} - {version_infos.version}')

def _create_logger(verbose, log_file):
    logger = logging.getLogger('ecs-crd')
    logger.setLevel(logging.INFO)
    formatter = None
    level = logging.DEBUG if verbose else logging.INFO
    logging.getLogger("boto3").setLevel(level)
    if level == logging.DEBUG:
        boto3.set_stream_logger(name='boto3', level=logging.DEBUG)
        boto3.set_stream_logger(name='botocore', level=logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s : %(name)s : %(levelname)s :  %(message)s')
    else:
        formatter = logging.Formatter('%(name)s : %(message)s')

    if log_file:
        logging.basicConfig(filename=log_file, level=logging.DEBUG)

    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import boto3

from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.canaryReleaseInfos import ScaleInfos
from ecs_crd.canaryReleaseInfos import SecretInfos
from ecs_crd.prepareDeploymentServiceDefinitionStep import PrepareDeploymentServiceDefinitionStep
from ecs_crd.updateCanaryReleaseInfoStep import UpdateCanaryReleaseInfoStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep


class PrepareDeploymentContainerDefinitionsStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(
            infos, f'Prepare {infos.action} ( Container definitions )', logger)

    def _process_container_name(self, source, target):
        """update the name informations for the current container"""
        self._process_property(
            source=source,
            target=target,
            source_property='name',
            multi=True,
            parent_property='Service.Container',
            default='default',
            indent=1
        )

    def _process_container_image(self, source, target):
        """update the image informations for the current container"""
        self._process_property(
            source=source,
            target=target,
            source_property='image',
            parent_property='Service.Container',
            default='{{account_id}}.dkr.ecr.{{region}}.amazonaws.com/{{name}}:{{version}}',
            indent=3
        )
        # check exist container image
        if self._is_container_image_from_ecr(source, target):
            if not self._exist_container_image_from_ecr(source, target):
                raise ValueError(
                    f'The container image {target["Image"]} is unknown in AWS ECR registry.')

    def _process_container_cpu(self, source, target):
        """update the cpu informations for the current container"""
        self._process_property(
            source=source,
            target=target,
            source_property='cpu',
            parent_property='Service.Container',
            type=int,
            default=128,
            indent=3
        )

    def _process_container_memory(self, source, target):
        """update the cpu informations for the current container"""
        self._process_property(
            source=source,
            target=target,
            source_property='memory',
            parent_property='Service.Container',
            type=int,
            default=128,
            indent=3
        )

    def _process_container_memory_reservation(self, source, target):
        """update the memory reservation informations for the current container"""
        self._process_property(
            source=source,
            target=target,
            source_property='memory_reservation',
            parent_property='Service.Container',
            type=int,
            indent=3
        )

    def _process_container_port_mappings(self, source, target):
        """update the port mappings informations for the current container"""
        if 'port_mappings' in source:
            target['PortMappings'] = []
            self._log_information(key='Port mappings', value='', indent=3)
            for p in source['port_mappings']:
                port_mapping = {
                    'HostPort': 0,
                    'ContainerPort': int(p['container_port'])
                }
                if 'host_port' in p:
                    if isinstance(p['host_port'], int):
                        port_mapping['HostPort'] = p['host_port']
                    elif 'green' in p['host_port'] and 'blue' in p['host_port']:
                        port_mapping['HostPort'] = int(
                            p['host_port'][self.infos.elected_release])
                protocol = 'tcp'
                host_port = ''
                if port_mapping['HostPort'] == 0:
                    host_port = '"dynamic"'
                else:
                    host_port = '{} ({})'.format(
                        port_mapping['HostPort'], self.infos.elected_release)
                if 'protocol' in p:
                    protocol = p['protocol']
                    port_mapping['Protocol'] = protocol
                target['PortMappings'].append(port_mapping)
                self._log_information(key='- '+protocol, value='{} -> {}'.format(
                    host_port, port_mapping['ContainerPort']), indent=4)

    def _process_container_entry_point(self, source, target):
        """update the entry point informations for the current container"""
        if 'entry_point' in source:
            if isinstance(source['entry_point'], list):
                target['EntryPoint'] = ','.join(source['entry_point'])
                self._log_information(
                    key='Entry point', value=target["EntryPoint"], indent=3)
            else:
                raise ValueError(
                    f'{source["entry_point"]} is not valid EntryPoint for Container.')

    def _process_container_environment(self, source, target):
        """update the environment informations for the current container"""
        target['Environment'] = []
        if 'environment' in source:
            for elmt in source['environment']:
                e = {}
                for k, v in elmt.items():
                    e['Name'] = k
                    e['Value'] = v
                e['Value'] = self._bind_data(str(v))
                target['Environment'].append(e)
        # ajout variable d'environement par défault
        self._add_default_environment_variable(
            target, 'AWS_ENVIRONMENT', '{{environment}}')
        self._add_default_environment_variable(
            target, 'AWS_REGION', '{{region}}')
        self._add_default_environment_variable(
            target, 'AWS_ACCOUNT_ID', '{{account_id}}')
        self._log_information(key='Environment', value='', indent=3)
        for e in target['Environment']:
            self._log_information(
                key=e['Name'], value=e['Value'], indent=4)

    def _process_container_secrets(self, source, target):
        """update the secrets informations for the current container"""
        target['Secrets'] = []
        if 'secrets' in source:
            self._log_information(key='Secrets', value='', indent=3)
            for elmt in source['secrets']:
                # TODO A retravailler
                e = {}
                for a in elmt.keys():
                    e['Name'] = a
                e['ValueFrom'] = None
                for a in elmt.values():
                    e['ValueFrom'] = a
                    e['ValueFrom'] = self._bind_data(e['ValueFrom'])
                for s in self.infos.secret_infos.secrets:
                    if s['id'] == e['ValueFrom']:
                        e['ValueFrom'] = s['arn']
                target['Secrets'].append(e)
                self._log_information(
                    key=e['Name'], value=e['ValueFrom'], indent=4)

    def _process_container_command(self, source, target):
        """update the command informations for the current container"""
        if 'command' in source:
            if isinstance(source['command'], list):
                self._log_information(key='Command', value='', indent=3)
                target['Command'] = []
                for e in source['command']:
                    target['Command'].append(e)
                    self._log_information(key='- '+e, value=None, indent=5)
            else:
                raise ValueError(
                    f'{source["command"]} is not valid Command for Container.')

    def _process_container_dns_search_domains(self, source, target):
        """update the dns search domain informations for the current container"""
        if 'dns_search_domains' in source:
            if isinstance(source['dns_search_domains'], list):
                self._log_information(
                    key='Dns Search Domains', value='', indent=3)
                target['DnsSearchDomains'] = []
                for e in source['dns_search_domains']:
                    target['DnsSearchDomains'].append(e)
                    self._log_information(key='- '+e, value=None, indent=5)
            else:
                raise ValueError(
                    f'{source["dns_search_domains"]} is not valid DnsSearchDomains for Container.')

    def _process_container_disable_networking(self, source, target):
        """update the disable networking informations for the current container"""
        self._process_property(
            source=source,
            target=target,
            source_property='disable_networking',
            parent_property='Container',
            type=bool,
            indent=3
        )

    def _process_container_dns_servers(self, source, target):
        """update the dns servers informations for the current container"""
        if 'dns_servers' in source:
            if isinstance(source['dns_servers'], list):
                self._log_information(key='Dns Servers', value='', indent=1)
                target['DnsServers'] = []
                for e in source['dns_servers']:
                    target['DnsServers'].append(e)
                    self._log_information(key='- '+e, value=None, indent=2)
            else:
                raise ValueError(
                    f'{source["dns_servers"]} is not valid DnsServers for Container.')

    def _process_container_links(self, item, container):
        """update the links informations for the current container"""
        if 'links' in item:
            self._log_information(key='Links', value='', indent=1)
            container['Links'] = []
            for e in item['links']:
                container['Links'].append(e)
                self._log_information(key='- '+e, value=None, indent=2)

    def _process_container_docker_security_options(self, item, container):
        """update the docker security informations for the current container"""
        if 'docker_security_options' in item:
            self._log_information(
                key='Docker Security Options', value='', ljust=10, indent=1)
            container['DockerSecurityOptions'] = []
            for e in item['docker_security_options']:
                container['DockerSecurityOptions'].append(e)
                self._log_information(key='- '+e, value=None, indent=2)

    def _process_container_essential(self, item, container):
        """update the essential informations for the current container"""
        container["Essential"] = "true"
        if 'essential' in item:
            if str(item['essential']).lower().strip() == 'true':
                container['Essential'] = "true"
            else:
                container['Essential'] = "false"
        self._log_information(
            key='Essential', value=container["Essential"], indent=3)

    def _is_container_image_from_ecr(self, item, container):
        return container['Image'].startswith(self._bind_data('{{account_id}}.dkr.ecr.{{region}}.amazonaws.com/'))

    def _exist_container_image_from_ecr(self, item, container):
        client = boto3.client('ecr', region_name=self.infos.region)
        try:
            data = container['Image'].split('/')
            registry = data[0]
            data = data[1].split(':')
            repository = data[0]
            tag = data[1]
            response = client.describe_images(
                repositoryName=repository, imageIds=[{"imageTag": tag}])
            return len(response["imageDetails"]) == 1
        except Exception as e:
            self.logger.error(e)
            return False

    def _process_container_privileged(self, item, container):
        """update the privileged informations for the current container"""
        if 'privileged' in item:
            if str(item['privileged']).lower().strip() == 'true':
                container['Privileged'] = "true"
            else:
                container['Privileged'] = "false"
            self._log_information(
                key='Priviliged', value=container["Privileged"], indent=1)
    
    def _process_container_healthcheck(self, item, container):
        """update the healthcheck informations for the current container"""
        if 'health_check' in item:
            self._log_information(key='HealthCheck', value='', indent=2)
            health_check = {}
            health_check['Command'] = []
            self._log_information(
                key='Command', value='', indent=3)
            for command in item['health_check']['command']:
                health_check['Command'].append(command)
                self._log_information(
                    key= '- ' +command, value=None, indent=4)                
            if 'interval' in item['health_check']:
                health_check['Interval'] = int(item['health_check']['interval'])
                self._log_information(
                    key='Interval', value=mount_point["Interval"], indent=2)
            if 'retries' in item['health_check']:
                health_check['Retries'] = int(item['health_check']['retries'])    
                self._log_information(
                    key='Retries', value=health_check["Retries"], indent=2)
            if 'start_period' in item['health_check']:
                health_check['StartPeriod'] = int(item['health_check']['start_period'])
                self._log_information(
                    key='StartPeriod', value=health_check["StartPeriod"], indent=2)
            if 'timeout' in item['health_check']:
                health_check['Timeout'] = int(item['health_check']['timeout'])   
                self._log_information(
                    key='Timeout', value=health_check["Timeout"], indent=2)         
            container['HealthCheck'] = health_check

    def _process_container_mount_points(self, item, container):
        """update the mount points informations for the current container"""
        if 'mount_points' in item:
            self._log_information(key='Mount points', value='', indent=1)
            container['MountPoints'] = []
            for e in item['mount_points']:
                mount_point = {}
                mount_point['ContainerPath'] = self._bind_data(
                    e['container_path'])
                self._log_information(
                    key='- Container Path', value=mount_point["ContainerPath"], indent=2)
                mount_point['SourceVolume'] = self._bind_data(
                    e['source_volume'])
                self._log_information(
                    key='  Source Volume', value=mount_point["SourceVolume"], indent=2)
                mount_point['ReadOnly'] = "false"
                if 'read_only' in e:
                    if str(e['read_only']).lower().strip() == 'true':
                        mount_point['ReadOnly'] = "true"
                    else:
                        mount_point['ReadOnly'] = "false"
                    self._log_information(
                        key='  Read Only', value=mount_point["ReadOnly"], indent=2)
                container['MountPoints'].append(mount_point)

    def _process_container_hostname(self, source, target):
        """update the hostname informations for the current container"""
        self._process_property(
            source=source,
            target=target,
            source_property='hostname',
            parent_property='Container',
            indent=3
        )

    def _process_container_start_timeout(self, source, target):
        """update the start timeout informations for the current container"""
        self._process_property(
            source=source,
            target=target,
            source_property='start_timeout',
            parent_property='Container',
            type=int,
            indent=3
        )

    def _process_container_stop_timeout(self, source, target):
        """update the stop timeout informations for the current container"""
        self._process_property(
            source=source,
            target=target,
            source_property='stop_timeout',
            parent_property='Container',
            type=int,
            indent=3
        )

    def _process_container_log_configuration(self, item, container):
        container['LogConfiguration'] = {}
        if 'log_configuration' in item:
            log_configuration = item['log_configuration']
            container['LogConfiguration']['LogDriver'] = log_configuration['log_driver']
            if 'options' in log_configuration:
                container['LogConfiguration']['Options'] = {}
                e = {}
                for k, v in log_configuration['options'].items():
                    e['Name'] = k
                    e['Value'] = self._bind_data(str(v))
                    container['LogConfiguration']['Options'].append(e)
            if 'secret_options' in log_configuration:
                container['LogConfiguration']['SecretOptions'] = {}
                for i in log_configuration['secret_options']:
                    e = {}
                    e['Name'] = i['name']
                    e['ValueFrom'] = self._bind_data(i['value_from'])
                    container['LogConfiguration']['SecretOptions'].append(e)                 

        else:
            container['LogConfiguration'] = {}
            container['LogConfiguration']['LogDriver'] = 'awslogs'
            container['LogConfiguration']['Options'] = {}
            container['LogConfiguration']['Options'][
                'awslogs-group'] = f'/aws/ecs/{self.infos.cluster_name}/service/{self.infos.service_name}'
            container['LogConfiguration']['Options']['awslogs-region'] = self.infos.region
            container['LogConfiguration']['Options']['awslogs-stream-prefix'] = self.infos.service_version

        self._log_information(key='Log Configuration', value='', indent=3)
        self._log_information(
            key='Log Driver', value = container['LogConfiguration']['LogDriver'], indent=4)
        self._log_information(key='Options', value='', indent=4)
        for k,v in container['LogConfiguration']['Options'].items():
            self._log_information(
                key='- Name', value= k, indent=5)
            self._log_information(
                key='  Value', value= v, indent=5)
        if 'SecretOptions' in container['LogConfiguration']:
            self._log_information(key='SecretOptions', value='', indent=4)
            for k, v in container['LogConfiguration']['SecretOptions'].items():
                self._log_information(
                    key='- Name', value=k, indent=5)
                self._log_information(
                    key='  ValueFrom', value=v, indent=5)                

    def _process_container_depends_on(self, item, container):
        """update the depends on for the current container"""
        if 'depends_on' in item:
            container['DependsOn'] = []
            self._log_information(key='Depend On', value='', indent=4)
            for i in item['depends_on']:
                depends_on = {}
                depends_on['Condition'] = i['condition']
                depends_on['ContainerName'] = i['container_name']
                container['DependsOn'].append(depends_on)
                self._log_information(
                    key='- ContainerName', value=depends_on['ContainerName'], indent=5)
                self._log_information(
                    key='  Condition', value=depends_on['Condition'], indent=5)

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            self.infos.secret_infos = self._find_secrets_task_informations()
            cfn = self.infos.green_infos.stack['Resources']['TaskDefinition']['Properties']['ContainerDefinitions']
            for source in self.configuration['service']['containers']:
                target = {}
                self._process_container_name(source, target)
                self._process_container_image(source, target)
                self._process_container_cpu(source, target)
                self._process_container_memory(source, target)
                self._process_container_memory_reservation(source, target)
                self._process_container_port_mappings(source, target)
                self._process_container_entry_point(source, target)
                self._process_container_environment(source, target)
                self._process_container_secrets(source, target)
                self._process_container_command(source, target)
                self._process_container_dns_search_domains(source, target)
                self._process_container_disable_networking(source, target)
                self._process_container_dns_servers(source, target)
                self._process_container_docker_security_options(source, target)
                self._process_container_essential(source, target)
                self._process_container_links(source, target)
                self._process_container_privileged(source, target)
                self._process_container_mount_points(source, target)
                self._process_container_hostname(source, target)
                self._process_container_log_configuration(source, target)
                self._process_container_start_timeout(source, target)
                self._process_container_stop_timeout(source, target)
                self._process_container_depends_on(source, target)
                self._process_container_healthcheck(source, target)

                cfn.append(target)
            self.infos.save()
            if self.infos.action == 'undeploy':
                return UpdateCanaryReleaseInfoStep(self.infos, self.logger)
            else:
                return PrepareDeploymentServiceDefinitionStep(self.infos, self.logger)

        except Exception as e:
            self.infos.exit_code = 4
            self.infos.exit_exception = e
            self.logger.error(self.title, exc_info=True)
            return SendNotificationBySnsStep(self.infos, self.logger)

    def _add_default_environment_variable(self, target, key, value):
        """add a new default environment variables if not exists"""
        if not any(filter(lambda x: x['Name'] == key, target['Environment'])):
            env = {}
            env['Name'] = key
            env['Value'] = self._bind_data(value)
            target['Environment'].append(env)

    def _find_secrets_task_informations(self):
        """find secret task information for vault resolution in excution task role."""
        result = None
        o = []

        for container_infos in self.configuration['service']['containers']:
            if 'secrets' in container_infos:
                for item in container_infos['secrets']:
                    format = self._bind_data(str(item)).replace('\'', '"')
                    j = json.loads(format)
                    o.append(j)
        # no secrets
        if not o:
            return None
        # secrets exist
        result = SecretInfos()
        client = boto3.client('secretsmanager', region_name=self.infos.region)
        kmsKeyIds = []
        for item in o:
            for k, v in item.items():
                secretId = v
                try:
                    response = client.describe_secret(SecretId=secretId)
                    if response['ARN'] not in result.secrets_arn:
                        a = {}
                        a['id'] = v
                        a['arn'] = response['ARN']
                        result.secrets.append(a)
                        result.secrets_arn.append(response['ARN'])
                        if response['KmsKeyId'] not in kmsKeyIds:
                            kmsKeyIds.append(response['KmsKeyId'])
                except Exception as e:
                    raise ValueError(f'Invalid secret: {secretId}, reason:{e}')

        client = boto3.client('kms',  region_name=self.infos.region)
        for k in kmsKeyIds:
            response = client.describe_key(
                KeyId=k, GrantTokens=['DescribeKey'])
            result.kms_arn.append(response['KeyMetadata']['Arn'])
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3

from ecs_crd.canaryReleaseInfos import ReleaseInfos
from ecs_crd.canaryReleaseInfos import LoadBalancerInfos
from ecs_crd.canaryReleaseInfos import ListenerRuleInfos
from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.prepareDeploymentScaleParametersStep import PrepareDeploymentScaleParametersStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep

class PrepareDeploymentLoadBalancerParametersStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos, f'Prepare {infos.action} ( Load Balancer parameters )', logger)

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            # récupération de l'item dynamo si il existe
            dynamodb_item = self._find_blue_dynamodb_item()

            # recherhcer de deux load balancers eligibles
            albs = self._find_load_balancers(dynamodb_item)

            # si le blue alb n'existe plus, on le supprime
            if dynamodb_item:
                if not any(filter(lambda x: x.arn == dynamodb_item['alb_arn'], albs)):
                    self._delete_obsolete_item()
                    dynamodb_item = None

            if len(albs) < 2:
                raise ValueError('There is not enough load balancer for a canary deployment.')
            elif len(albs) > 2:
                raise ValueError('There are too many load balancers for a canary deployment.')

            self.infos.init_infos.stack['Parameters']['LoadBalancerBlue']['Default'] = albs[0].dns_name
            self.infos.init_infos.stack['Parameters']['LoadBalancerGreen']['Default'] = albs[1].dns_name
            self.infos.init_infos.stack_name = self._generate_name(canary_release='0')

            exist = self._find_cloud_formation_stack(self.infos.init_infos.stack_name)

            if exist:
                 self.infos.init_infos.stack_id = exist['StackId']
                 self.infos.init_infos.stack = None

            # green infos
            greens = list(filter(lambda x: x.is_elected, albs))
            if not greens:
                raise ValueError(f'There is no ALB to deploy the Green release.')
            green = greens[0]

            self.infos.green_infos.stack['Parameters']['LoadBalancer']['Default'] = green.arn
            self.infos.green_infos.stack['Parameters']['CanaryRelease']['Default'] = green.canary_release
            self.infos.green_infos.alb_arn = green.arn
            self.infos.green_infos.canary_release = green.canary_release
            self.infos.green_infos.alb_dns = green.dns_name
            self.infos.green_infos.alb_hosted_zone_id = green.hosted_zone_id

            self.infos.green_infos.stack_name = self._generate_name(canary_release=green.canary_release)
            exist = self._find_cloud_formation_stack(self.infos.green_infos.stack_name)
            if exist:
                raise ValueError(f'There is already a cloudformation stack named for the Green release : {self.infos.green_infos.stack_name}.') 

            # init infos
            stack = self._find_cloud_formation_stack(self.infos.init_infos.stack_name)
            if stack and self.infos.action=='deploy':
                # no recreate init stack for deploy
                self.infos.init_infos.stack = None
            
            self.infos.save()

            # blue infos
            blues = list(filter(lambda x: x != green, albs))
            blue = blues[0]
            self.infos.blue_infos = ReleaseInfos()
            self.infos.blue_infos.alb_arn = blue.arn
            self.infos.blue_infos.alb_dns = blue.dns_name
            self.infos.blue_infos.canary_release = blue.canary_release
            self.infos.blue_infos.alb_hosted_zone_id = green.hosted_zone_id
            if dynamodb_item:
                exist_blue_deployment = self._find_cloud_formation_stack(dynamodb_item['stack_name'])
                if exist_blue_deployment:
                    self.infos.blue_infos.stack_id = exist_blue_deployment['StackId']
                    self.infos.blue_infos.stack_name = f'{self.infos.environment}-{self.infos.service_name}-{blue.canary_release}'
            
            self._log_sub_title('Load balancer "blue" {}'.format('(elected)' if self.infos.elected_release == 'blue' and self.infos.action=='deploy' else ''))
            self._log_information(key='Fqdn', value=blue.dns_name, ljust=4)
            self._log_information(key='Arn', value=blue.arn, ljust=4)

            self._log_sub_title('Load balancer "green" {}'.format('(elected)' if self.infos.elected_release == 'green' and self.infos.action=='deploy' else ''))
            self._log_information(key='Fqdn', value=green.dns_name, indent=1, ljust=4)
            self._log_information(key='Dns', value=green.arn, indent=1, ljust=4)

            self.infos.save()
            return PrepareDeploymentScaleParametersStep(self.infos, self.logger)

        except Exception as e:
            self.infos.exit_code = 2
            self.infos.exit_exception = e
            self.logger.error(self.title, exc_info=True)
            return SendNotificationBySnsStep(self.infos, self.logger)
         
    def _find_cloud_formation_stack(self, stack_name, next_token=None):
        client = boto3.client('cloudformation', region_name=self.infos.region)
        response = None
        if next_token:
            response = client.list_stacks(StackStatusFilter=['CREATE_COMPLETE'], NextToken=next_token )
        else:
            response = client.list_stacks(StackStatusFilter=['CREATE_COMPLETE'])
        count = sum(1 for e in response['StackSummaries'] if e['StackName'] == stack_name)
        if count == 0:
            if 'NextToken' in response:
                return self._find_cloud_formation_stack(stack_name, response['NextToken'])
            else:
                return None
        response = client.describe_stacks(StackName=stack_name)
        if len(response['Stacks']) == 1:
            return response['Stacks'][0]
        else:
            return None

    def _find_load_balancers(self, dynamodb_item):
        client = boto3.client('elbv2', region_name=self.infos.region)
        response = client.describe_load_balancers()
        albs = []
        for item in response['LoadBalancers']:
            if item['VpcId'] != self.infos.vpc_id or item['Type'] != 'application':
                continue
            arn = item['LoadBalancerArn']
            dnsName = item['DNSName']
            canonicalHostedZoneId = item['CanonicalHostedZoneId']
            canaryRelease = None
            canaryGroup = None
            # find tag informations
            response = client.describe_tags(ResourceArns=[arn])
            for tagDescriptions in response['TagDescriptions']:
                for tag in tagDescriptions['Tags']:
                    if tag['Key'] == 'CanaryRelease':
                        canaryRelease = tag['Value']
                    if tag['Key'] == 'CanaryGroup' and tag['Value'] == self.infos.canary_group:
                        canaryGroup = self.infos.canary_group

            # aln is ready 
            if canaryRelease and canaryGroup:
                albs.append(
                    LoadBalancerInfos(
                        arn=arn, 
                        dns_name= dnsName , 
                        canary_release = canaryRelease, 
                        hosted_zone_id=canonicalHostedZoneId
                    )
                )

        # no alb found to deploy
        if not albs:
            raise ValueError(f'There are no tagged ALBs ("{self.infos.environment}") allowing the service to be deployed.')

        blue_alb_arn = None
        if dynamodb_item:
            blue_alb_arn = dynamodb_item['alb_arn']

        elected_alb = None

        # rechecher de l'alb green
        for alb in albs:
            if alb.arn != blue_alb_arn:
                alb.is_elected = True
                elected_alb = alb
                response = client.describe_listeners(LoadBalancerArn=alb.arn)
                for item in self.configuration['listeners']:
                    # rechercher de l'écouteur associé au port 
                    listeners = list(filter(lambda x: x['Port'] == int(item['port']),response['Listeners']))
                    # le port est déjà écouté sur le 'load balancer' , on doit créer un 'listener rule'
                    if len(listeners) == 1:
                        listener = listeners[0]
                        # on vérifie qu'il y a une règle pour le
                        if 'rules' not in item:
                            container_name = 'default'
                            if 'container' in item['target_group'] and 'name' in item['target_group']['container']:
                                container_name = item['target_group']['container']['name']
                            raise ValueError('The listener port {} is already used for Alb:{}. You can either use a different port for the listener or set rules on the listener. ( ContainerName:{}, ContainerPort:{} )'.format(item['port'], alb.arn, container_name,  item['target_group']['container']['port'], ))
                        rule_infos = ListenerRuleInfos(listener_arn=listener['ListenerArn'], configuration=item)
                        self.infos.listener_rules_infos.append(rule_infos)
                break

        if not elected_alb:
            raise ValueError('Not Load balencer found for deploy.')
        self.infos.elected_release = 'blue' if str(elected_alb.canary_release) == str(self.configuration['canary']['releases']['blue']) else 'green'
        return albs

    def _delete_obsolete_item(self):
        client = boto3.resource('dynamodb', region_name=self.infos.region)
        table = client.Table('canary_release')
        table.delete_item(Key={'id': self.infos.get_hash()})

    def _find_blue_dynamodb_item(self):
        client = boto3.resource('dynamodb', region_name = self.infos.region)
        table = client.Table('canary_release')
        response = table.get_item(Key={'id': self.infos.get_hash()})

        if 'Item' in response:
            return response['Item']
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ecs_crd.canaryReleaseInfos import PolicyInfos
from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.prepareDeploymentStrategyStep import PrepareDeploymentStrategyStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep

class PrepareDeploymentIamPoliciesStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos, f'Prepare {infos.action} ( IAM Role & Policies )', logger)

    def _convert_to_dto_policy_info(self, item, count):
        name = f'policy_{count}'
        if 'name' in item:
            name = item['name']
        resources = []
        effect = "Allow"
        if 'effect' in item:
            effect = item['effect']
        if 'resources' not in item:
            resources.append('*')
        else:
            for r in item['resources']:
                resources.append(self._bind_data(r))
        actions = item['actions']
        return PolicyInfos(
            name = name, 
            effect =effect, 
            actions = actions, 
            resources =resources)

    def _find_all_task_role_policies(self):
        """find all task role policies policies """
        result = []
        if 'iam_roles' in self.configuration["service"] and 'task_role' in self.configuration["service"]['iam_roles']:
            count = 1
            for item in self.configuration['service']['iam_roles']['task_role']:
                result.append(self._convert_to_dto_policy_info(item, count))
        return result

    def _find_all_task_execution_role_policies(self):
        """find all task execution role policies """
        result = []
        if 'iam_roles' in self.configuration["service"] and 'task_execution_role' in self.configuration["service"]['iam_roles']:
            count = 1
            for item in self.configuration['service']['iam_roles']['task_execution_role']:
                result.append(self._convert_to_dto_policy_info(item, count))
        
        # add secret policy
        if self.infos.secret_infos:
            cfn_policies = []
            effect = 'Allow'
            actions = ['kms:Decrypt', 'secretsmanager:GetSecretValue']
            resources = self.infos.secret_infos.secrets_arn + self.infos.secret_infos.kms_arn
            result.append(PolicyInfos(
                name = 'AllowReadSecrets', 
                effect = effect, 
                actions = actions, 
                resources = resources ))
        return result

    def _policy_info_2_cloud_formation_policy(self, policy_info):
        """convert a dto policy to cloud formation IAM policy"""
        cfn_policy = {}
        cfn_policy['PolicyName'] = policy_info.name
        cfn_policy['PolicyDocument'] = {}
        cfn_policy['PolicyDocument']['Version'] = "2012-10-17"
        cfn_policy['PolicyDocument']['Statement'] = []
        cfn_policy_item = {}
        cfn_policy_item['Effect'] = policy_info.effect
        cfn_policy_item['Action'] = policy_info.actions
        cfn_policy_item['Resource'] = policy_info.resources
        cfn_policy['PolicyDocument']['Statement'].append(cfn_policy_item)
        self.logger.info('')
        self._log_information(key='Sid', value=policy_info.name, indent=1)
        self._log_information(key='Effect', value=policy_info.effect, indent=2)
        self._log_information(key='Action', value='', indent=2)
        for a in policy_info.actions:
            self._log_information(key='- '+a, value=None, indent=3)
        self._log_information(key='Resource', value='', indent=2)
        for a in policy_info.resources:
            self._log_information(key='- '+a, value=None, indent=3)
        return cfn_policy

    def _process_task_role(self):
        """update the role for the ECS task service"""
        policies = self._find_all_task_role_policies()
        if policies:
            #TODO Revoir la creation du HASH
            self._log_sub_title('Task role definition')
            self.logger.info('')
            role = {}
            role['Type'] = 'AWS::IAM::Role'
            role['Properties'] = {}
            role['Properties']['RoleName'] = self._generate_name(suffix='-ecs-task', canary_release=self.infos.green_infos.canary_release)
            self._log_information(key='Name', value=role['Properties']['RoleName'])
            role['Properties']['AssumeRolePolicyDocument'] = {}
            role['Properties']['AssumeRolePolicyDocument']['Version'] = '2012-10-17'
            role['Properties']['AssumeRolePolicyDocument']['Statement'] = []
            item = {}
            item['Effect'] = 'Allow'
            item['Principal'] = {}
            item['Principal']['Service'] = []
            item['Principal']['Service'].append('ecs-tasks.amazonaws.com')
            item['Action'] = []
            item['Action'].append('sts:AssumeRole')
            role['Properties']['AssumeRolePolicyDocument']['Statement'].append(item)
            cfn_policies = []
            policy_infos = self._find_all_task_role_policies()
            for policy_info in policy_infos:
                cfn_policies.append(self._policy_info_2_cloud_formation_policy(policy_info))
            role['Properties']['Policies'] = cfn_policies
            self.infos.green_infos.stack['Resources']['TaskRole'] = role
            self.infos.green_infos.stack['Resources']['TaskDefinition']['Properties']['TaskRoleArn'] = {}
            self.infos.green_infos.stack['Resources']['TaskDefinition']['Properties']['TaskRoleArn']['Ref'] = 'TaskRole'

    def _process_task_execution_role(self):
        """update the role for ECS task execution"""
        self.logger.info('')
        self._log_sub_title('Task execution role definition')
        self.logger.info('')
        self.infos.green_infos.stack['Resources']['TaskExecutionRole']['Properties']['RoleName'] = self._generate_name(suffix='-ecs-exec-task', canary_release=self.infos.green_infos.canary_release)
        self._log_information(key='Name', value=self.infos.green_infos.stack['Resources']['TaskExecutionRole']['Properties']['RoleName'])
        policy_infos = self._find_all_task_execution_role_policies()
        if policy_infos:
            cfn_policies = []
            for policy_info in policy_infos:
                cfn_policies.append(self._policy_info_2_cloud_formation_policy(policy_info))
            self.infos.green_infos.stack['Resources']['TaskExecutionRole']['Properties']['Policies'] = cfn_policies

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            self._process_task_role()
            self._process_task_execution_role()
            self.infos.save()
            return PrepareDeploymentStrategyStep(self.infos, self.logger)

        except Exception as e:
            self.infos.exit_code = 9
            self.infos.exit_exception = e
            self.logger.error(self.title, exc_info=True)
            return SendNotificationBySnsStep(self.infos, self.logger)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3
import time
import uuid
import datetime

from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep

class FinishDeploymentStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos, f'Finish {infos.action}', logger, with_end_log=False)

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        if self.infos.exit_code == 0:
            self.logger.info('Result      : COMPLETED')
        else:
            self.logger.info('Result      : FAILED')
            self.logger.info(f'Exit code   : {self.infos.exit_code}')
            self.logger.info(f'Exit error  : {self.infos.exit_exception}')
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3
import urllib.request
import json

from ecs_crd.canaryReleaseInfos import StrategyInfos
from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.prepareDeploymentLoadBalancerParametersStep import PrepareDeploymentLoadBalancerParametersStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep
from ecs_crd.canaryReleaseInfos import FqdnInfos

class PrepareDeploymentGlobalParametersStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos, f'Prepare {infos.action} ( Global parameters )', logger)

    def _process_account_id(self):
        """update the AWS account ID informations for the service"""
        self.infos.account_id = boto3.client('sts').get_caller_identity().get('Account')
        self._log_information(key='Account ID', value=self.infos.account_id, ljust=18)

    def _process_env_data(self):
        self.infos.green_infos.stack['Parameters']['Environment']['Default'] = self.infos.environment
        self.infos.green_infos.stack['Parameters']['Region']['Default'] = self.infos.region
        self.infos.init_infos.stack['Parameters']['Environment']['Default'] = self.infos.environment
        self.infos.init_infos.stack['Parameters']['Region']['Default'] = self.infos.region

    def _process_canary_group(self):
        """update the canary group name informations for the service"""
        self.infos.canary_group = self.configuration['canary']['group']
        self._log_information(key='Canary group', value=self.infos.canary_group, ljust=18)

    def _process_external_ip(self):
        """update the external ip informations for the service"""
        self.infos.external_ip = self._find_external_ip()
        self._log_information(key='External IP', value=self.infos.external_ip, ljust=18)

    def _process_project(self):
        """update the project name informations for the service"""
        self.infos.project = self.configuration['service']['project']
        self._log_information(key='Project', value=self.infos.project, ljust=18)
        self.infos.green_infos.stack['Parameters']['ProjectName']['Default'] = self.infos.project
        self.infos.init_infos.stack['Parameters']['ProjectName']['Default'] = self.infos.project

    def _process_service_name(self):
        """update the name informations for the service"""
        self.infos.service_name = self._bind_data(self.configuration['service']['name'])
        self.infos.green_infos.stack['Parameters']['ServiceName']['Default'] = self.infos.service_name
        self.infos.init_infos.stack['Parameters']['ServiceName']['Default'] = self.infos.service_name
        self._log_information(key='Service', value=self.infos.service_name, ljust=18)

    def _process_version(self):
        """update the version informations for the service"""
        version = 'latest'
        if 'version' in self.configuration['service']:
            version = str(self.configuration['service']['version'])
        self.infos.service_version = version
        self.infos.green_infos.stack['Parameters']['Version']['Default'] = self.infos.service_version
        self._log_information(key='Version', value=self.infos.service_version, ljust=18)

    def _process_vpc_id(self):
        """update the AWS vpc ID informations for the service"""
        self.infos.vpc_id = self._find_vpc_Id()
        self._log_information(key='Vpc ID', value=self.infos.vpc_id, ljust=18)

    def _process_fqdn(self):
        """update the fqdn"""

        
        source = self.configuration['service']['fqdn']
        if isinstance(source, str):
            self.infos.fqdn.append(self._to_fqdn_infos(source))
        if isinstance(source, list):
            for item in source:
                self.infos.fqdn.append(self._to_fqdn_infos(item))
        self._log_information(key='Fqdn', value=self.infos.fqdn[0].name, ljust=18)

        if len(self.infos.fqdn) > 1:
            for i in range(1, len(self.infos.fqdn)):
                self.logger.info('                    '+self.infos.fqdn[i].name)
 
    def _process_cluster(self):
        """update the AWS ECS cluster informations for the service"""
        # clusterName 
        clusterName = self._bind_data('{{environment}}-ecs-cluster')
        if 'cluster' in self.configuration['service']:
            clusterName = self._bind_data(self.configuration['service']['cluster'])
        self.infos.cluster_name = clusterName
        self._log_information(key='Cluster', value=self.infos.cluster_name, ljust=18)
        self.infos.green_infos.stack['Parameters']['ClusterName']['Default'] = clusterName
        self.infos.init_infos.stack['Parameters']['ClusterName']['Default'] = clusterName

        # cluster
        cluster = self._find_cluster(clusterName)
        self.infos.cluster = cluster
        self.infos.green_infos.stack['Parameters']['Cluster']['Default'] = cluster
        self._log_information(key='Cluster ID', value=self.infos.cluster, ljust=18)

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            self._log_information(key=f'{self.infos.action} ID', value=self.infos.id, ljust=18)
            self._process_account_id()
            self._process_env_data()
            self._process_canary_group()
            self._process_external_ip()
            self._process_project()
            self._process_service_name()
            self._process_version()
            self._process_vpc_id()
            self._process_fqdn()
            self._process_cluster()
            self.infos.save()
            self._create_dynamodb_table()
            return PrepareDeploymentLoadBalancerParametersStep(self.infos, self.logger)
        except Exception as e:
            self.infos.exit_code = 1
            self.infos.exit_exception = e
            self.logger.error(self.title, exc_info=True)
            return SendNotificationBySnsStep(self.infos, self.logger)
        else:
            return None

    def _find_vpc_Id(self):
        """find the AWS VPC by environment"""
        ec2 = boto3.resource('ec2', region_name=self.infos.region)
        client = boto3.client('ec2', region_name=self.infos.region)
        ids = map(lambda x: x.id, list(ec2.vpcs.filter(Filters=[])))
        for id in ids:
            response = client.describe_vpcs(VpcIds=[id])
            if 'Tags' in response['Vpcs'][0]:
                for tag in response['Vpcs'][0]['Tags']:
                    if tag['Key'] == 'Environment' and tag['Value'] == self.infos.environment:
                        return id
        raise ValueError('vpc id {} not found for environment'.format(self.infos.environment))

    def _find_cluster(self, clusterName):
        """find the AWS ECS cluster by name"""
        client = boto3.client('ecs', region_name=self.infos.region)
        response = client.list_clusters()
        for arn in response['clusterArns']:
            if arn.endswith(clusterName):
                return arn
        raise ValueError(f'Cluster "{clusterName}" not found.')

    def _create_dynamodb_table(self):
        """create the AWS DynamoDB table if not exist"""
        client = boto3.client('dynamodb', region_name=self.infos.region)
        table_name = 'canary_release'
        existing_tables = client.list_tables()['TableNames']
        if table_name not in existing_tables:
            client.create_table(
                AttributeDefinitions=[
                    {
                        'AttributeName': 'id',
                        'AttributeType': 'S',
                    },
                ],
                KeySchema=[
                    {
                        'AttributeName':'id',
                        'KeyType': 'HASH',
                    },
                ],
                ProvisionedThroughput={
                    'ReadCapacityUnits': 2,
                    'WriteCapacityUnits': 2,
                },
                TableName='canary_release',
            )

    def _find_external_ip(self):
        """find the external ip"""
        data = None
        try:
            data = json.loads(urllib.request.urlopen("https://api.ipify.org?format=json").read())
            if 'ip' in data:
                return data['ip']
        except Exception:
            pass
        return data


    def _to_fqdn_infos(self, source):
        name = self._bind_data(source)
        data = name.strip('.').split('.')
        hosted_zone_name = data[len(data)-2]+'.'+data[len(data)-1]
        hosted_zone = self._find_hosted_zone(hosted_zone_name.strip('.')+'.')
        hosted_zone_id = hosted_zone['Id']
        result = FqdnInfos(
            name = name, 
            hosted_zone_name = hosted_zone_name,
            hosted_zone_id = hosted_zone_id
        )
        return result

    def _find_hosted_zone(self, hostZoneName):
        """find the AWS Route53 dns zone by name"""
        client = boto3.client('route53', region_name=self.infos.region)
        response = client.list_hosted_zones()
        for item in response['HostedZones']:
            if item['Name'] == hostZoneName:
                return item



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3
import time
from abc import ABC, abstractmethod

from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep

class DestroyStackStep(CanaryReleaseDeployStep):

    def __init__(self, infos, title, logger, stack_infos ):
        """initializes a new instance of the class"""
        super().__init__(
            infos, 
            title, 
            logger
        )
        self.timer = 5
        self.stack_infos = stack_infos

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            if self.stack_infos.stack_id:
                client = boto3.client('cloudformation', region_name=self.infos.region)
                self._destroy_stack(client)
                self._monitor(client)
            else:
                self.logger.info('Not destruction stack (reason: the stack not exist).')
            self.stack_infos.stack_id = None   
            self.infos.save()
            return self._on_success()
        except Exception as e:
            self.infos.exit_exception = e
            self.infos.exit_code = 17
            self.logger.error(self.title, exc_info=True)
            return self._on_fail()
   
    @abstractmethod
    def _on_success(self):
        pass
        
    @abstractmethod
    def _on_fail(self):
        pass

    def _destroy_stack(self, client):
        """destroys the cloud formation stack"""
        client.delete_stack(StackName=self.stack_infos.stack_id)

    def _monitor(self, client):
        """pause the process and wait for the result of the cloud formation stack deletion"""
        wait = 0
        while True:
            wait += self.timer
            w = self._second_to_string(wait)
            self.logger.info('')
            time.sleep(self.timer)
            self.logger.info(f'Deleting stack in progress ... [{w} elapsed]')
            response = client.describe_stacks(StackName=self.stack_infos.stack_id)
            stack = response['Stacks'][0]
            response2 = client.list_stack_resources(StackName=self.stack_infos.stack_id)
            for resource in response2['StackResourceSummaries']:
                message = resource['LogicalResourceId'].ljust(40, '.') + resource['ResourceStatus']
                if 'ResourceStatusReason' in resource:
                    message += f' ( {resource["ResourceStatusReason"]} )'
                self.logger.info(message)

            if stack['StackStatus'] == 'DELETE_IN_PROGRESS':
                continue
            else:
                if stack['StackStatus'] == 'DELETE_COMPLETE':
                    break
                else:
                    raise ValueError('Error deletion cloudformation stack')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
from ecs_crd.canaryReleaseDeployStep import CanaryReleaseDeployStep
from ecs_crd.prepareDeploymentTaskDefinitionStep import PrepareDeploymentTaskDefinitionStep
from ecs_crd.sendNotificationBySnsStep import SendNotificationBySnsStep

class PrepareDeploymentServiceDefinitionStep(CanaryReleaseDeployStep):

    def __init__(self, infos, logger):
        """initializes a new instance of the class"""
        super().__init__(infos, f'Prepare {infos.action} ( Service definition )', logger)

    def _process_scheduling_strategy(self, source, target):
        """update the sceduling stratégy informations for the service"""
        self._process_property(
            source = source,
            target = target,
            pattern= 'DAEMON|REPLICA',
            source_property = 'scheduling_strategy',
            parent_property = 'Service',
            indent = 1
        )
    
    def _process_platform_version(self, source, target):
        """update the plaform version informations for the service"""
        self._process_property(
            source = source,
            target = target,
            source_property = 'platform_version',
            parent_property = 'Service',
            indent = 1
        )
  
    def _process_placement_constraints(self, source, target):
        """update the placement constraintes informations for the service"""
        if 'placement_constraints' in source:
            self._log_information(key='Placement Contraints', value='',indent=1)
            placement_constraints = []
            for item in source['placement_constraints']:
                constraint = {}
                self._process_placement_constraints_contraint_type(item, constraint)
                self._process_placement_constraints_contraint_expression(item, constraint)
                placement_constraints.append(constraint)
            target['PlacementConstraints'] = placement_constraints

    def _process_placement_constraints_contraint_type(self, source, target):
        self._process_property(
            source = source,
            target = target,
            pattern= 'distinctInstance|memberOf',
            multi = True,
            required = True,
            source_property = 'type',
            parent_property = 'Service.PlacementConstraints.Type',
            indent = 2
        )

    def _process_placement_constraints_contraint_expression(self, source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'expression',
            parent_property = 'Service.PlacementConstraints.Expression',
            indent = 4
        )
    
    def _process_placement_stategies(self, source, properties):
        """update the placement strategies informations for the service"""
        if 'placement_strategies' in source:
            self._log_information(key='Placement Strategies',value='', ljust=10, indent=1)
            placement_stategies = []
            for item in source['placement_strategies']:
                strategy = {}
                self._process_placement_stategies_strategy_type(item, strategy)
                self._process_placement_stategies_strategy_field(item,strategy)
                placement_stategies.append(strategy)
            properties['PlacementStrategies'] = placement_stategies
    
    def _process_placement_stategies_strategy_field(self,source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'field',
            required = True,
            multi = False,
            parent_property = 'Service.PlacementStrategies',
            indent = 4
        )

    def _process_placement_stategies_strategy_type(self,source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'type',
            required = True,
            pattern = 'binpack|random|spread',
            multi = True,
            parent_property = 'Service.PlacementStrategies',
            indent = 2
        )

    def _process_load_balancer(self):
        cfn = self.infos.green_infos.stack['Resources']['Service']['Properties']['LoadBalancers']
        for item in self.configuration['service']['containers']:
            if 'port_mappings' in item:
                for e in item['port_mappings']:
                    definition = {}
                    container_name = 'default'
                    if 'name' in item:
                        container_name = item['name']
                    definition['ContainerName'] =  container_name
                    definition['ContainerPort'] = int(e['container_port'])
                    definition['TargetGroupArn'] = {}
                    w = definition['ContainerName'] + '_' + str(e['container_port'])
                    definition['TargetGroupArn']['Ref'] = "TargetGroup{}".format(''.join(id.capitalize() for id in w.split('_')))
                    cfn.append(definition)

    def _process_auto_scaling(self, source):            
        if 'auto_scaling' in source:
            self._log_information(key = "AutoScaling", value='', indent=1)
            source = source['auto_scaling']
            self._process_application_autoscaling_scalable_target(source)
            count = 1
            for item in source['auto_scaling_policies']:
                self._process_application_auto_scaling_scaling_policy(item, count)

    def _process_application_autoscaling_scalable_target_min_capacity(self, source, target):
       self._process_property(
            source = source,
            target = target,
            source_property = 'min_capacity',
            type = int,
            default = self.infos.scale_infos.desired,
            parent_property = 'Service.AutoScaling',
            indent = 2
        )

    def _process_application_autoscaling_scalable_target_max_capacity(self, source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'max_capacity',
            type = int,
            default = self.infos.scale_infos.desired,
            parent_property = 'Service.AutoScaling',
            indent = 2
        )

    def _process_application_autoscaling_scalable_target_role_arn(self, source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'role_arn',
            target_property = 'RoleARN',
            required = True,
            parent_property = 'Service.AutoScaling',
            indent = 2
        )

    def _process_application_autoscaling_scalable_target(self, source):
        cfn = {}
        cfn['Type'] = 'AWS::ApplicationAutoScaling::ScalableTarget'
        target = {}
        # min_capacity
        self._process_application_autoscaling_scalable_target_min_capacity(source, target)
        # max_capacity
        self._process_application_autoscaling_scalable_target_max_capacity(source, target)
        # role_arn
        self._process_application_autoscaling_scalable_target_role_arn(source, target)
        target['ResourceId'] = f'service/{self.infos.cluster_name}/{self.infos.service_name}-{self.infos.green_infos.canary_release}'
        target['ScalableDimension'] = 'ecs:service:DesiredCount'
        target['ServiceNamespace'] = 'ecs'
        cfn['Properties'] = target
        cfn['DependsOn'] = 'Service'
        self.infos.green_infos.stack['Resources']['AutoScalingTarget'] = cfn

    def _process_application_auto_scaling_scaling_policy_policy_name(self, source, target, count):
        self._process_property(
            source = source,
            target = target,
            source_property = 'policy_name',
            multi = True,
            parent_property = 'Service.AutoScaling',
            default = f'{self.infos.service_name}-scaling-policy-{count}',
            indent = 3
        )

    def _process_application_auto_scaling_scaling_policy_policy_type(self, source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'policy_type',
            pattern ='SimpleScaling|StepScaling|TargetTrackingScaling',
            parent_property = 'Service.AutoScaling',
            default = 'StepScaling',
            indent = 5
        )

    def _process_application_auto_scaling_scaling_policy(self, source, count):
        cfn = {}
        self._log_information(key = 'AutoScalingPolicies', value= '', indent=2)
        cfn['Type'] = 'AWS::ApplicationAutoScaling::ScalingPolicy'
        target = {}
        # policy_name
        self._process_application_auto_scaling_scaling_policy_policy_name(source, target, count)
        # policy_type
        self._process_application_auto_scaling_scaling_policy_policy_type(source, target)
        target['ScalingTargetId'] =  {}
        target['ScalingTargetId']['Ref'] = 'AutoScalingTarget'
        target['ScalableDimension'] = 'ecs:service:DesiredCount'
        target['ServiceNamespace'] = 'ecs'
        
        self._process_step_scaling_policy_configuration(source, target)
        self._process_cloudwatch_alarms(source, target, count)
        cfn['Properties'] = target
        self.infos.green_infos.stack['Resources'][f'AutoScalingPolicy{count}'] = cfn

    def _process_step_scaling_policy_configuration_adjustment_type(self, source, target):
        """process step_scaling_policy_configuration.adjustment_type"""
        self._process_property(
            source = source,
            target = target,
            source_property = 'adjustment_type',
            pattern = 'ChangeInCapacity|ExactCapacity|PercentChangeInCapacity',
            default = 'ChangeInCapacity',
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration',
            indent = 6
        )
       
    def _process_step_scaling_policy_configuration_cooldown(self, source, target):
        """process step_scaling_policy_configuration.cooldown"""
        self._process_property(
            source = source,
            target = target,
            source_property = 'cooldown',
            type = int,
            default = 60,
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration',
            indent = 6
        )
   
    def _process_step_scaling_policy_configuration_metric_aggregation_type(self, source, target):
        """process step_scaling_policy_configuration.metric_aggregation_type"""
        self._process_property(
            source = source,
            target = target,
            source_property = 'metric_aggregation_type',
            pattern = 'Average|Minimum|Maximum',
            default = 'Average',
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration',
            indent = 6
        )
      
    def _process_step_scaling_policy_configuration(self, source, target):
        if 'step_scaling_policy_configuration' not in source:
            raise ValueError('The StepScalingPolicyConfiguration is required')
        sub_source = source['step_scaling_policy_configuration']   
        sub_target = {}
        self._log_information(key = 'StepScalingPolicyConfiguration', value='', indent=5)
        # adjustment_type
        self._process_step_scaling_policy_configuration_adjustment_type(sub_source, sub_target)

        # cooldown
        self._process_step_scaling_policy_configuration_cooldown(sub_source,sub_target)
        # metric_aggregation_type
        self._process_step_scaling_policy_configuration_metric_aggregation_type(sub_source, sub_target)
        # step_adjustments
        self._process_step_scaling_policy_configuration_step_adjustments(sub_source, sub_target)
        target['StepScalingPolicyConfiguration'] = sub_target

    def _process_step_scaling_policy_configuration_step_adjustments(self, source, target):
        if 'step_adjustments' not in source:
            raise ValueError('StepAdjustments is required')
        target['StepAdjustments'] = []
        self._log_information(key='StepAdjustments',value ='',indent=6)
        for sub_source in source['step_adjustments']:
                sub_target = {}
                # metric_interval_lower_bound
                self._process_step_scaling_policy_configuration_step_adjustments_metric_interval_lower_bound(sub_source, sub_target)
                # metric_interval_upper_bound
                self._process_step_scaling_policy_configuration_step_adjustments_metric_interval_upper_bound(sub_source, sub_target)
                # scaling_adjustment
                self._process_step_scaling_policy_configuration_step_adjustments_scaling_adjustment(sub_source, sub_target)
                target['StepAdjustments'].append(sub_target)
   
    def _process_step_scaling_policy_configuration_step_adjustments_metric_interval_lower_bound(self, source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'metric_interval_lower_bound',
            type = int,
            multi = True,
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration.StepAdjustment',
            indent = 7
        )

    def _process_step_scaling_policy_configuration_step_adjustments_metric_interval_upper_bound(self, source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'metric_interval_upper_bound',
            type = int,
            multi = True,
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration.StepAdjustment',
            indent = 7
        )

    def _process_step_scaling_policy_configuration_step_adjustments_scaling_adjustment(self, source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'scaling_adjustment',
            type = int,
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration.StepAdjustment',
            indent = 9
        )

    def _process_cloudwatch_alarms(self, source, target, count):
        if 'cloudwatch_alarms' not in source:
            raise ValueError('CloudwatchAlarms is required for Service.AutoScaling.StepScalingPolicyConfiguration.')
        self._log_information(key='CloudWatch Alarms', value='', indent=5)
        count_cloudwatch_alarms = 1
        for sub_source in source['cloudwatch_alarms']:
            self._process_cloudwatch_alarm(sub_source, count, count_cloudwatch_alarms)
            count_cloudwatch_alarms = count_cloudwatch_alarms + 1

    def _process_cloudwatch_alarm_metric_name(self, source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'metric_name',
            pattern = 'CPUUtilization|MemoryUtilization',
            default = 'Average',
            multi = True,
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration.CloudwatchAlarms',
            indent = 6
        )

    def _process_cloudwatch_alarm_alarm_description(self, source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'alarm_description',
            default = f'Containers {target["MetricName"]} High',
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration.CloudwatchAlarms',
            indent = 8
        )

    def _process_cloudwatch_alarm_namespace(self, source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'namespace',
            default = 'AWS/ECS',
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration.CloudwatchAlarms',
            indent = 8
        )

    def _process_cloudwatch_alarm_statistic(self, source, target):
        self._process_property(
            source = source,
            target = target,
            source_property = 'statistic',
            pattern = 'Average|Maximum|Minimum|SampleCount|Sum',
            default = 'Average',
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration',
            indent = 8
        )
    
    def _process_cloudwatch_alarm_period(self, source, target):
       self._process_property(
            source = source,
            target = target,
            source_property = 'period',
            type = int,
            default = 300,
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration',
            indent = 8
        )

    def _process_cloudwatch_alarm_evaluation_periods(self, source, target):
       self._process_property(
            source = source,
            target = target,
            source_property = 'evaluation_periods',
            type = int,
            default = 1,
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration',
            indent = 8
        )
 
    def _process_cloudwatch_alarm_threshold(self, source, target):
       self._process_property(
            source = source,
            target = target,
            source_property = 'threshold',
            type = int,
            required = True,
            parent_property = 'Service.AutoScaling.StepScalingPolicyConfiguration',
            indent = 8
        )

    def _process_cloudwatch_alarm_comparison_operator(self, origin, target):
        if 'comparison_operator' in origin:
            val = re.match('GreaterThanOrEqualToThreshold|GreaterThanThreshold|LessThanOrEqualToThreshold|LessThanThreshold', origin['comparison_operator'])
            if not val:
                raise ValueError(f'{origin["comparison_operator"]} is not valid for CloudwatchAlarm.ComparisonOperator.')
            target['ComparisonOperator'] = val.group(0)
        else:
           raise ValueError('ComparisonOperator is required in CloudwatchAlarm')
        self._log_information(key = 'ComparisonOperator', value=target['ComparisonOperator'], indent=8)

    def _process_cloudwatch_alarm(self, origin, count_policy, count_cloudwatch_alarms):
        cfn = {}
        cfn['Type'] = 'AWS::CloudWatch::Alarm'
        target = {}
        # metric_name
        self._process_cloudwatch_alarm_metric_name(origin, target)
        # alarm_description
        self._process_cloudwatch_alarm_alarm_description(origin, target)
        # namespace
        self._process_cloudwatch_alarm_namespace(origin, target)
        # statistic
        self._process_cloudwatch_alarm_statistic(origin, target)
        # period
        self._process_cloudwatch_alarm_period(origin, target)
        # evaluation_periods
        self._process_cloudwatch_alarm_evaluation_periods(origin, target)
        # threshold
        self._process_cloudwatch_alarm_threshold(origin, target)
        # comparison_operator
        self._process_cloudwatch_alarm_comparison_operator(origin, target)

        target['AlarmActions'] = []
        alarm_action = {}
        alarm_action['Ref'] = f'AutoScalingPolicy{count_policy}'
        target['AlarmActions'].append(alarm_action)

        dimensions = []
        dimension = {}
        dimension['Name'] = 'ServiceName'
        dimension['Value'] = f'{self.infos.service_name}-{self.infos.green_infos.canary_release}'
        dimensions.append(dimension)
        dimension = {}
        dimension['Name'] = 'ClusterName'
        dimension['Value'] = f'{self.infos.cluster_name}'
        dimensions.append(dimension)
        target['Dimensions'] = dimensions

        self._log_information(key = '  Dimensions', value='', indent=6)
        for d in target['Dimensions']:
            self._log_information(key = '- Name', value=d['Name'], indent=9)
            self._log_information(key = 'Value', value=d['Value'], indent=11)
        cfn['Properties'] = target
        cfn['DependsOn'] = f'AutoScalingPolicy{count_policy}'
        self.infos.green_infos.stack['Resources'][f'AutoScalingAlarm{count_cloudwatch_alarms}'] = cfn

    def _on_execute(self):
        """operation containing the processing performed by this step"""
        try:
            source = self.configuration['service']
            target = self.infos.green_infos.stack['Resources']['Service']['Properties']

            self._process_scheduling_strategy(source, target)
            self._process_platform_version(source, target)
            self._process_placement_constraints(source, target)
            self._process_placement_stategies(source, target)
            self._process_load_balancer()
            self._process_auto_scaling(source)
            self.infos.save()
            return PrepareDeploymentTaskDefinitionStep(self.infos, self.logger)
        except Exception as e:
            self.infos.exit_code = 5
            self.infos.exit_exception = e
            self.logger.error(self.title, exc_info=True)
            return SendNotificationBySnsStep(self.infos, self.logger)
